# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import sys
import atexit
import base64
import hashlib
from math import ceil
from datetime import datetime, timezone

from ..fs import FsPath


def timestamp(*, in_filename=False, ms=False, us=False):
    """Timestamp for easy use in logs and filenames.
    Args:
        ms (bool): If True, milliseconds are included.
            Default False.
        in_filename (bool): If True, colons are replaced
            with dashes to be used in filenames. Default False.
    Returns:
        str: Current timestamp in UTC.
    """
    if ms and us:
        raise ValueError("Only one of 'ms' or 'us' can be True!")
    idx = -3 if ms else -7
    idx = 26 if us else idx
    form = "%Y-%m-%d_%H-%M-%S.%f" if in_filename else "%Y-%m-%d %H:%M:%S.%f"
    return datetime.now(tz=timezone.utc).strftime(form)[:idx]


def ranID(*, length=12, size=1, only_alphanumeric=False):
    """Base64 encoded random ID.
    Args:
        length (int): Length of the ID string, rounded up to
            the closest multiple of 4. Default 12.
        size (int): Number of random IDs to generate.
            Default 1.
        only_alphanumeric (bool): If True, only alphanumeric
            characters are used. Default False.
    Returns:
        str: Random ID string.
    """
    if length < 1:
        raise ValueError("Length must be greater than 0!")
    if size < 1:
        raise ValueError("Size must be greater than 0!")
    if size > 1:
        return [ranID(length=length, only_alphanumeric=only_alphanumeric)
                for _ in range(size)]
    length = int(ceil(length/4))
    if only_alphanumeric:
        ran = ''
        for _ in range(length):
            while True:
                this_ran = ranID(length=4, only_alphanumeric=False)
                if this_ran.isalnum():
                    break
            ran += this_ran
        return ran
    else:
        random_bytes = os.urandom(3*length)
        return base64.urlsafe_b64encode(random_bytes).decode('utf-8')


def system_lock(lockfile):
    """Create a lockfile or quit the process if it already exists.
    This is useful for cronjobs that might overlap in time if they
    run for too long.
    Args:
        lockfile (str): Path to the lockfile.
    """
    lockfile = FsPath(lockfile)
    # Check if previous process still running
    if lockfile.exists():
        sys.exit(f"Previous {lockfile.name} script still active! Exiting...")
    else:
        # Otherwise register a lockfile
        lockfile.touch()
        def exit_handler():
            lockfile.unlink()
        atexit.register(exit_handler)


def get_hash(filename, *, size=128):
    """Get a fast hash of a file, in chunks of 'size' (in kb)"""
    h  = hashlib.blake2b()
    b  = bytearray(size*1024)
    mv = memoryview(b)
    with open(filename, 'rb', buffering=0) as f:
        for n in iter(lambda : f.readinto(mv), 0):
            h.update(mv[:n])
    return h.hexdigest()



# copyright ############################### #
# This file is part of the Xaux package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import sys
from .gh import *
from .package_manager import get_latest_package_version


class VersionError(OSError):
    pass


def make_release_branch(package, bump=None, allow_major=False, ignore_name=False):
    if bump is None:
        bump, _, ignore_name = _parse_argv(optional_force=False)

    # Check necessary setup and installs
    assert_git_repo()
    if not ignore_name:
        assert_git_repo_name(package)
    _assert_in_root_package_dir(package)
    assert_poetry_installed()

    # Check our working directory is clean
    git_assert_working_tree_clean()
    branch = git_current_branch()
    if branch != "main":
        raise GitError("This script needs to be ran on the main branch.")
    git_pull()   # Sync with the remote to be sure we don't delete an incomplete branch later
    git_push()
    print("Repository is clean.")

    expected_ver = poetry_get_expected_version(bump)
    _assert_not_major_version(expected_ver, 'make_release_branch.py', allow_major)

    branch = f"release/v{expected_ver}"
    expected_ver = f"{expected_ver}rc0"
    _confirm_version_bump(expected_ver)

    print("Creating release branch...")
    git_switch(branch, create=True)
    print("Poetry version bump...")
    new_ver = _do_bump(expected_ver)

    _adapt_version_files(package, new_ver)
    git_add(["pyproject.toml", f"{package}/general.py", "tests/test_version.py"])
    git_commit(f"Created release branch release/v{new_ver}.", no_verify=True)
    git_push(set_upstream=True)

    print("All done!")


def rename_release_branch(package, bump=None, allow_major=False, ignore_name=False):
    if bump is None:
        bump, _, ignore_name = _parse_argv(optional_force=False)

    # Check necessary setup and installs
    assert_git_repo()
    if not ignore_name:
        assert_git_repo_name(package)
    _assert_in_root_package_dir(package)
    assert_poetry_installed()

    # Check our working directory is clean
    git_assert_working_tree_clean()
    branch = git_current_branch()
    current_ver = poetry_get_version()
    if branch != f"release/v{current_ver[:-3]}":
        raise GitError("This script needs to be ran from a release branch.")
    git_pull()   # Sync with the remote to be sure we don't delete an incomplete branch later
    git_push()
    _assert_no_open_prs(branch)
    print("Repository is clean.")

    expected_ver = poetry_get_expected_version(bump)
    _assert_not_major_version(expected_ver, 'rename_release_branch.py', allow_major)

    new_branch = f"release/v{expected_ver}"
    expected_ver = f"{expected_ver}rc0"
    _confirm_version_bump(expected_ver)

    print("Renaming release branch...")
    git_rename_current_branch(new_branch, set_upstream=True)
    print("Poetry version bump...")
    new_ver = _do_bump(expected_ver)

    _adapt_version_files(package, new_ver)
    git_add(["pyproject.toml", f"{package}/general.py", "tests/test_version.py"])
    git_commit(f"Renamed release branch {branch} into {new_branch}.", no_verify=True)
    git_push(set_upstream=True)

    print("All done!")


def make_release(package, bump=None, force=False, allow_major=False, ignore_name=False):
    if bump is None:
        bump, force, ignore_name = _parse_argv(optional_force=True)

    # Check necessary setup and installs
    assert_git_repo()
    if not ignore_name:
        assert_git_repo_name(package)
    _assert_in_root_package_dir(package)
    assert_poetry_installed()
    assert_gh_installed()

    # Check our working directory is clean
    print("Verifying repository status...")
    git_assert_working_tree_clean()
    branch = git_current_branch()
    if branch == "main":
        raise GitError("\nThis script cannot be ran on the main branch."
                        "Make a release branch and make the new release from there."
                        "Make sure that the release branch has an upstream version (i.e."
                        "push at least once before running this script), or this script"
                        "will fail.")
    expected_ver = poetry_get_expected_version(bump)
    if not force:
        if branch != f"release/v{expected_ver}":
            raise VersionError(f"\nYou are bumping to {expected_ver} but this branch is {branch}. "
                                    "If this is intentional, use --force.")
    git_pull()   # Sync with the remote to be sure we don't delete an incomplete branch later
    git_push()
    _assert_no_open_prs(branch)
    print("Repository is clean.")

    _assert_not_major_version(expected_ver, 'release.py', allow_major)
    _confirm_version_bump(expected_ver)
    print("Updating version in the release branch...")
    new_ver = _do_bump(expected_ver, bump)

    _adapt_version_files(package, new_ver)
    _set_dependencies(package)
    git_add(["pyproject.toml", f"{package}/general.py", "tests/test_version.py"])
    git_commit(f"Updated version number to v{new_ver}.", no_verify=True)
    git_push()

    print("Creating and merging pull request to main branch...")
    gh_pr_create('main', f"Release {new_ver}")
    git_switch('main')
    git_pull()
    prs = gh_pr_list(base='main', head=branch)
    if len(prs) != 1:
        raise GitError(f"Expected one PR from {branch} to main, found {len(prs)}:\n"
                    + "\n".join([f"PR#{pr} from {br}" for pr, br in prs.items()]))
    gh_pr_merge(list(prs.keys())[0], admin=True, delete_branch=True)
    git_pull()
    git_make_tag(f"v{new_ver}")

    print("Creating draft release and publishing to PyPi...")
    gh_release_create(f"v{new_ver}", f"{package.capitalize()} release {new_ver}", draft=True)
    poetry_publish(build=True)

    print("All done!")


def _parse_argv(optional_force=False):
    # Check the script arguments
    options = []
    bump = None
    error = False
    for arg in sys.argv[1:]:
        if arg.startswith('--'):
            options.append(arg[2:])
        else:
            if bump is not None:
                error = True
                break
            bump = arg
    if error or bump is None:
        raise ValueError("Are you running CLI?\nThen this script needs exactly one argument: "
                       + "the new version number or a bump (which can be: patch, minor, major).\n"
                       + "If running in python, please provide the argument `bump=...`.\n"
                      + f"Provided arguments are: {sys.argv}")
    force = False
    ignore_name = False
    for opt in options:
        if opt == 'force':
            if not optional_force:
                raise ValueError("Cannot use option '--force'")
            force = True
        elif opt == 'ignore-name':
            ignore_name = True
        else:
            raise ValueError("Unknown option '{opt}'")
    return bump, force, ignore_name


def _assert_in_root_package_dir(package):
    if package in [pp.name for pp in Path.cwd().parents]:
        raise VersionError("This script needs to be ran from the root package directory.")


def _assert_not_major_version(expected_ver, file, allow_major):
    # Check that we are not accidentally bumping a major version
    if not allow_major:
        major_ver = int(expected_ver.split('.')[0])
        if major_ver != 0:
            raise VersionError("Bumping a major version! If this is really what you want, "
                            + f"then adapt {file} manually to add `allow_major=True`.")


def _assert_no_open_prs(branch):
    # Check that there are no conflicting PRs open
    prs = gh_pr_list(base=branch)
    if prs:
        raise GitError(f"There are open PRs to the release branch:\n" \
                    + "\n".join([f"PR#{pr} from {br}" for pr, br in prs.items()]) \
                    + "\nThese would be automatically closed by this script, " \
                    + "as the target branch disappears. Please close them manually, " \
                    + "or change the target branch.")
    prs = gh_pr_list(base='main', head=branch)
    if prs:
        raise GitError(f"There are open PRs from the release branch to main:\n" \
                    + "\n".join([f"PR#{pr} from {br}" for pr, br in prs.items()]) \
                    + "\nThese would conflict with the versioning script. "
                    + "Please close them manually.")

def _set_dependencies(package):
    # Manually get the xsuite dependencies from the pyproject.toml file (not from PyPi as things might have changed)
    xsuite_pkgs = ['xaux', 'xobjects', 'xdeps', 'xtrack', 'xpart', 'xfields', 'xcoll', 'xdyna', 'xboinc', 'xwakes']
    xsuite_pkgs.remove(package)
    latest_version = {}
    for pkg in xsuite_pkgs:
        latest_version[pkg] = get_latest_package_version(pkg)
    with Path("pyproject.toml").open("r") as fid:
        lines = fid.readlines()
    with Path("pyproject.toml").open("w") as fid:
        for line in lines:
            if any([pkg in line and ">=" in line for pkg in xsuite_pkgs]):
                for pkg in xsuite_pkgs:
                    if pkg in line and ">=" in line:
                        # Bit of hackery to support both "xtrack>=0.89.3", and xaux = ">=0.3.5"
                        parts = line.split('>=')
                        if len(parts) != 2:
                            raise VersionError(f"Fatal error: could not parse line {line}...")
                        parts = [parts[0], *parts[1].split('"')]
                        fid.write('"'.join([f"{parts[0]}>={latest_version[pkg]}", *parts[2:]]))
                        break
            else:
                fid.write(line)

def _confirm_version_bump(expected_ver):
    current_ver = poetry_get_version()
    print(f"Bumping from {current_ver} to {expected_ver}.")
    print("Type y to continue (or anything else to cancel):")
    answer = input()
    if answer not in ["y", "Y"]:
        print("Cancelled.")
        sys.exit(1)


def _do_bump(expected_ver, bump=None):
    if bump is None:
        bump=expected_ver
    poetry_bump_version(bump)
    new_ver = poetry_get_version()
    if new_ver != expected_ver:
        raise VersionError(f"Fatal error: `poetry --dry-run` expected {expected_ver}, but result is {new_ver}..."
                            "Need to recover manually!")
    return new_ver


def _adapt_version_files(package, new_ver):
    for file, pattern in zip([f"{package}/general.py", "tests/test_version.py"],
                            ["__version__ = ", "    assert __version__ == "]):
        file_adapted = False
        with Path(file).open("r") as fid:
            lines = fid.readlines()
        with Path(file).open("w") as fid:
            for line in lines:
                if line.startswith(pattern):
                    fid.write(f"{pattern}'{new_ver}'\n")
                    file_adapted = True
                else:
                    fid.write(line)
        if not file_adapted:
            raise VersionError(f"Fatal error: could not adapt {file}...")



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os
from math import prod
from subprocess import run, PIPE, CalledProcessError


_xrdcp_installed = False
if os.name != 'nt':
    try:
        cmd = run(["xrdcp", "--version"], stdout=PIPE, stderr=PIPE)
        _xrdcp_installed = cmd.returncode == 0
    except (CalledProcessError, FileNotFoundError):
        _xrdcp_installed = False


# The class os.stat_result can only be initialised with a Tuple,
# however, the contents of this Tuple can change between python versions.
# This is a rather contrived way of finding the Tuple indices that correspond
# to certain fields.
_stat_n_fields = os.stat_result.n_fields
_stat_idx = {k: [] for k in os.stat_result.__dict__ if k.startswith('st_')}
for i in range(os.stat_result.n_sequence_fields, _stat_n_fields):
    _stat_def = os.stat_result(tuple(range(i)))
    for k in _stat_def.__dir__():
        if k.startswith('st_'):
            val = getattr(_stat_def, k, None)
            if val is not None and val not in _stat_idx[k]:
                _stat_idx[k] += [val]
# Sanity check; if this fails, that certain indices are used for different fields
assert len(sum(_stat_idx.values(), [])) == len(set(sum(_stat_idx.values(), [])))

# Make a stat_result from the fields directly
def make_stat_result(stat_dict):
    stats = [None for _ in range(_stat_n_fields)]
    for key, val in stat_dict.items():
        if key in _stat_idx:
            for st in _stat_idx[key]:
                stats[st] = int(val)
    return os.stat_result(tuple(stats))


def size_expand(size, binary=False):
    size = str(size).lower()
    size = size.replace('b', '')
    if binary:
        size = size.replace('k','*1024')
        size = size.replace('m', '*1048576')
        size = size.replace('g', '*1073741824')
        size = size.replace('t', '*1099511627776')
    else:
        size = size.replace('k','*1000')
        size = size.replace('m', '*1000000')
        size = size.replace('g', '*1000000000')
        size = size.replace('t', '*1000000000000')
    return int(prod([int(i) for i in size.split('*')]))




# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import tempfile

_tempdir = tempfile.TemporaryDirectory()



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os, sys
from subprocess import run, PIPE, CalledProcessError
from time import sleep
from shutil import rmtree
from contextlib import contextmanager
from pathlib import Path, PurePosixPath, PureWindowsPath


def _non_strict_resolve(path, _as_posix=False):
    # This function is resolving without strict=True as this 
    # might lead to infinite loops in case of broken links.
    if not isinstance(path, Path):
        path = Path(path)
    cls = path.__class__
    path = path.expanduser().as_posix()
    try:
        path = os.path.realpath(path, strict=False)
    except TypeError:
        path = os.path.realpath(path)
    if _as_posix:
        return path
    else:
        return cls(path)


class FsPath:
    """Factory that generates either an EosPath, AfsPath, or LocalPath
    depending on the file system the path is on. Note that a LocalPath
    is just a regular Path object but with access to the FsPath methods.
    """
    __slots__ = ()
    __hash__ = Path.__hash__

    def __new__(cls, *args):
        from .eos import EosPath, _on_eos
        from .afs import AfsPath, _on_afs
        if len(args) == 0:
            args = ('.',)
        if _on_eos(*args):
            return EosPath.__new__(EosPath, *args, _eos_checked=True)
        elif _on_afs(*args):
            return AfsPath.__new__(AfsPath, *args, _afs_checked=True)
        else:
            return LocalPath.__new__(LocalPath, *args)

    # FsPath is not a subclass of Path. We get the instance methods
    # from Path via the derived classes (EosPath etc), but we have to
    # define the public class methods manually.

    @classmethod
    def cwd(cls):
        return cls(Path.cwd())

    @classmethod
    def home(cls):
        return cls(Path.home())


    # PurePath methods
    # ================

    if sys.version_info >= (3, 12):
        def with_segments(self, *pathsegments, _cast_as_fspath=True, _force_path=False):
            if _cast_as_fspath:
                return FsPath(*pathsegments)
            elif _force_path:
                return Path(*pathsegments)
            else:
                return type(self)(*pathsegments)

        # We create a contextmanager to avoid casting to FsPath in the constructor
        @classmethod
        @contextmanager
        def _in_constructor(cls, _force=False):
            original_with_segments = cls.with_segments
            def new_with_segments(self, *args, **kwargs):
                return original_with_segments(self, *args, _cast_as_fspath=False,
                                              _force_path=_force, **kwargs)
            cls.with_segments = new_with_segments
            try:
                yield
            finally:
                cls.with_segments = original_with_segments
                # cls.parent = original_parent

    else:
        @classmethod
        def _from_parts(cls, args, _cast_as_fspath=True, _force_path=False, **kwargs):
            if _cast_as_fspath:
                return FsPath(super(Path, cls)._from_parts(args, **kwargs))
            elif _force_path:
                return Path(super(Path, cls)._from_parts(args, **kwargs))
            else:
                return super(Path, cls)._from_parts(args, **kwargs)

        @classmethod
        def _from_parsed_parts(cls, *args, _cast_as_fspath=True, _force_path=False, **kwargs):
            if _cast_as_fspath:
                return FsPath(super(Path, cls)._from_parsed_parts(*args, **kwargs))
            elif _force_path:
                return Path(super(Path, cls)._from_parsed_parts(*args, **kwargs))
            else:
                return super(Path, cls)._from_parsed_parts(*args, **kwargs)

        # We create a contextmanager to avoid casting to FsPath in the constructor
        @classmethod
        @contextmanager
        def _in_constructor(cls, _force=False):
            original_from_parts = cls._from_parts
            @classmethod
            def new_from_parts(cls, args, **kwargs):
                return original_from_parts(args, _cast_as_fspath=False,
                                           _force_path=_force, **kwargs)
            cls._from_parts = new_from_parts
            original_from_parsed_parts = cls._from_parsed_parts
            @classmethod
            def new_from_parsed_parts(cls, *args, **kwargs):
                return original_from_parsed_parts(*args, _cast_as_fspath=False,
                                                  _force_path=_force, **kwargs)
            cls._from_parsed_parts = new_from_parsed_parts
            try:
                yield
            finally:
                cls._from_parts = original_from_parts
                cls._from_parsed_parts = original_from_parsed_parts


    # Overwrite Path methods
    # ======================

    # Resolving EOS paths can be tricky due to different mount points.
    # Luckily, the path is already resolved at instantiation.
    def resolve(self, *args, **kwargs):
        # We first resolve all internal symlinks
        from .eos import EosPath
        if isinstance(self, EosPath):
            new_path = FsPath(_non_strict_resolve(Path(self.eos_path), _as_posix=True))
        else:
            new_path = FsPath(Path.resolve(Path(self).expanduser()))
        # And then we get back the correct EOS path
        # This extra step is needed because the final path might not be on EOS
        if isinstance(new_path, EosPath):
            return EosPath(new_path.eos_path)
        else:
            return new_path

    if sys.version_info >= (3, 12):
        def exists(self, *args, follow_symlinks=True, **kwargs):
            if self.is_symlink(*args, **kwargs) and follow_symlinks:
                return self.resolve(*args, **kwargs).exists(*args, **kwargs)
            return Path.exists(self.expanduser(), follow_symlinks=follow_symlinks)

    else:
        def exists(self, *args, **kwargs):
            if self.is_symlink(*args, **kwargs):
                return self.resolve(*args, **kwargs).exists(*args, **kwargs)
            return Path.exists(self.expanduser())

    if sys.version_info >= (3, 13):
        def is_file(self, *args, follow_symlinks=True, **kwargs):
            return Path.is_file(self.expanduser(), follow_symlinks=follow_symlinks)

        def is_dir(self, *args, follow_symlinks=True, **kwargs):
            return Path.is_dir(self.expanduser(), follow_symlinks=follow_symlinks)

    else:
        def is_file(self, *args, **kwargs):
            return Path.is_file(self.expanduser())

        def is_dir(self, *args, **kwargs):
            return Path.is_dir(self.expanduser())

    def is_symlink(self, *args, **kwargs):
        return Path.is_symlink(self.expanduser())

    def symlink_to(self, target, target_is_directory=False, **kwargs):
        target = FsPath(target)
        return Path.symlink_to(self.expanduser().resolve(**kwargs), target.expanduser(),
                               target_is_directory=target.is_dir(**kwargs))

    def unlink(self, missing_ok=False, **kwargs):
        if not self.is_symlink(**kwargs) and self.is_dir(**kwargs):
            raise IsADirectoryError(f"{self} is a directory.")
        Path.unlink(self.expanduser(), missing_ok=missing_ok)

    def rmdir(self, *args, **kwargs):
        if not self.is_dir(*args, **kwargs):
            raise NotADirectoryError(f"{self} is not a directory.")
        Path.rmdir(self.expanduser())

    def __eq__(self, other):
        try:
            other = FsPath(other).expanduser().resolve()
        except:
            return False
        self = self.expanduser().resolve()
        return self.as_posix() == other.as_posix()

    def __ne__(self, other):
        return not self.__eq__(other)


    # New methods
    # ===========

    def getfid(self):
        cmd = run(['ls', '-i', self.expanduser().as_posix()],
                    stdout=PIPE, stderr=PIPE)
        if cmd.returncode == 0:
            stdout = cmd.stdout.decode('UTF-8').strip()
            return int(stdout.split()[0])
        return -1

    def flush(self):
        cmd = run(['sync', self.expanduser().as_posix()],
                    stdout=PIPE, stderr=PIPE)
        if cmd.returncode:
            sleep(1)
        self.touch()

    def lexists(self, *args, **kwargs):
        return self.is_symlink(*args, **kwargs) or self.exists(*args, **kwargs)

    def is_broken_symlink(self, *args, **kwargs):
        return self.is_symlink(*args, **kwargs) and not self.exists(*args, **kwargs)

    def rmtree(self, *args, **kwargs):
        if not self.is_dir(*args, **kwargs):
            raise NotADirectoryError(f"{self} is not a directory.")
        return rmtree(self.expanduser().resolve(*args, **kwargs).as_posix(*args, **kwargs), *args, **kwargs)

    def copy_to(self, dst, recursive=None, *args, **kwargs):
        from .io import cp
        if recursive is None:
            recursive = self.is_dir(*args, **kwargs)
        return cp(self, dst, *args, recursive=recursive, **kwargs)

    def move_to(self, dst, *args, **kwargs):
        from .io import mv
        return mv(self, dst, *args, **kwargs)

    def size(self, *args, **kwargs):
        return self.expanduser().stat(*args, **kwargs).st_size


# To give regular Path objects the same functionality as FsPath objects
class LocalPath(FsPath, Path):
    """Path subclass for local paths.

    Instantiating an FsPath should call this class.
    """

    def __new__(cls, *args):
        if cls is LocalPath:
            cls = LocalWindowsPath if os.name == 'nt' else LocalPosixPath
        with cls._in_constructor():
            try:
                self = cls._from_parts(args)
            except AttributeError:
                self = Path.__new__(cls, *args)
        return self

    def __init__(self, *args):
        with self.__class__._in_constructor():
            if sys.version_info >= (3, 12):
                Path.__init__(self, *args)
            else:
                Path.__init__(self)


class LocalPosixPath(LocalPath, PurePosixPath):
    """LocalPath subclass for local paths on non-Windows systems.

    On a POSIX system, instantiating an FsPath or a LocalPath
    should return this object.
    """
    __slots__ = ()

    if os.name == 'nt':
        def __new__(cls, *args, **kwargs):
            raise OSError(
                f"Cannot instantiate {cls.__name__!r} on your system")


class LocalWindowsPath(LocalPath, PureWindowsPath):
    """LocalPath subclass for local paths on Windows systems.

    On a Windows system, instantiating an FsPath or a LocalPath
    should return this object.
    """
    __slots__ = ()

    if os.name != 'nt':
        def __new__(cls, *args, **kwargs):
            raise OSError(
                f"Cannot instantiate {cls.__name__!r} on your system")




# copyright ############################### #
# This file is part of the Xaux package.    #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import inspect


def count_arguments(func, count_variable_length_args=False):
    i = 0
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.POSITIONAL_ONLY \
        or param.kind == inspect.Parameter.KEYWORD_ONLY \
        or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            i += 1
        if count_variable_length_args:
            if param.kind == inspect.Parameter.VAR_POSITIONAL \
            or param.kind == inspect.Parameter.VAR_KEYWORD:
                i += 1
    return i

def count_required_arguments(func):
    i = 0
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if (param.kind == inspect.Parameter.POSITIONAL_ONLY \
        or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD) \
        and param.default == inspect.Parameter.empty:
            i += 1
    return i

def count_optional_arguments(func):
    i = 0
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if (param.kind == inspect.Parameter.KEYWORD_ONLY \
        or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD) \
        and param.default != inspect.Parameter.empty:
            i += 1
    return i

def has_variable_length_arguments(func):
    return has_variable_length_positional_arguments(func) \
           or has_variable_length_keyword_arguments(func)

def has_variable_length_positional_arguments(func):
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            return True
    return False

def has_variable_length_keyword_arguments(func):
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .fs import FsPath, LocalPath, LocalPosixPath, LocalWindowsPath
from .afs import AfsPath, AfsPosixPath, AfsWindowsPath, afs_accessible
from .eos import EosPath, EosPosixPath, EosWindowsPath
from .eos_methods import eos_accessible, is_egroup_member
from .fs_methods import make_stat_result, size_expand
from .io import cp, mv

_xrdcp_use_ipv4 = True

# For testing
_skip_afs_software = False
_skip_eos_software = False
_force_xrdcp = False
_force_eoscmd = False



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import sys
import json
from pathlib import Path
from subprocess import run

_GIT_REPO = False
_GH_INSTALLED = False
_POETRY_INSTALLED = False


def assert_git_repo():
    global _GIT_REPO
    if not _GIT_REPO:
        try:
            cmd = run(["git", "status"], capture_output=True)
            if cmd.returncode != 0:
                raise GitError(f"{Path.cwd()} is not a git repository.")
            else:
                _GIT_REPO = True
        except FileNotFoundError:
            raise GitError("git is not installed.")

def assert_git_repo_name(repo):
    try:
        cmd = run(["git", "rev-parse", "--show-toplevel"], capture_output=True)
        if cmd.returncode != 0:
            raise GitError(f"{Path.cwd()} is not a git repository.")
        else:
            if Path(cmd.stdout.decode('UTF-8').strip()).name != repo:
                raise GitError(f"{Path.cwd()} is not in the {repo} repository.")
    except FileNotFoundError:
        raise GitError("git is not installed.")

def assert_gh_installed():
    global _GH_INSTALLED
    if not _GH_INSTALLED:
        try:
            cmd = run(["gh", "--version"], capture_output=True)
            if cmd.returncode != 0:
                raise GhError("gh is not installed.")
            else:
                _GH_INSTALLED = True
        except FileNotFoundError:
            raise GhError("gh is not installed.")

def assert_poetry_installed():
    global _POETRY_INSTALLED
    if not _POETRY_INSTALLED:
        try:
            cmd = run(["poetry", "--version"], capture_output=True)
            if cmd.returncode != 0:
                raise PoetryError("poetry is not installed.")
            else:
                _POETRY_INSTALLED = True
        except FileNotFoundError:
            raise PoetryError("poetry is not installed.")


def git_assert_working_tree_clean():
    assert_git_repo()
    cmd = run(["git", "diff", "--quiet"], capture_output=True)
    if cmd.returncode != 0:
        raise GitError("Some changes are not staged. Stage and commit first.")
    cmd = run(["git", "diff", "--staged", "--quiet"], capture_output=True)
    if cmd.returncode != 0:
        raise GitError("Some changes are staged. Commit first.")

def _run_git(cmds):
    assert_git_repo()
    cmd = run(["git", *cmds], capture_output=True)
    if cmd.returncode == 0:
        return cmd.stdout.decode('UTF-8').strip()
    else:
        raise GitError(' '.join(cmds) + ":\n" + cmd.stderr.decode('UTF-8').strip())

def git_current_branch():
    return _run_git(["symbolic-ref", "--short", "HEAD"])

def git_rename_current_branch(new_branch, set_upstream=False):
    old_branch = git_current_branch()
    if old_branch == 'main':
        raise GitError("Cannot rename the main branch.")
    out = _run_git(["branch", "-m", new_branch])
    if out: print(out)
    if set_upstream:
        out = _run_git(["push", "origin", "--delete", old_branch])
        if out: print(out)
        git_push(set_upstream=True)

def git_switch(branch, create=False):
    cc = ['-c'] if create else []
    out = _run_git(["switch", *cc, branch])
    if out: print(out)

def git_add(files):
    out = _run_git(["add", *files])
    if out: print(out)

def git_commit(message, no_verify=False):
    vv = ['--no-verify'] if no_verify else []
    out = _run_git(["commit", *vv, '-m', message])
    if out: print(out)

def git_pull():
    out = _run_git(["pull"])
    if out: print(out)

def git_push(set_upstream=False):
    uu = ["--set-upstream", "origin", git_current_branch()] if set_upstream else []
    out = _run_git(["push", *uu])
    if out: print(out)

def git_make_tag(tag):
    out = _run_git(["tag", tag])
    if out: print(out)
    out = _run_git(["push", "origin", tag])
    if out: print(out)


def _run_gh(cmds):
    assert_gh_installed()
    cmd = run(["gh", *cmds], capture_output=True)
    if cmd.returncode == 0:
        return cmd.stdout.decode('UTF-8').strip()
    else:
        raise GhError(' '.join(cmds) + ":\n" + cmd.stderr.decode('UTF-8').strip())

def gh_pr_create(base_branch, title):
    out = _run_gh(["pr", "create", "--base", base_branch, "--title", title, '--fill'])
    if out: print(out)

def gh_pr_list(base=None, head=None):
    bb = [] if base is None else ["-B", base]
    hh = [] if head is None else ["-H", head]
    out = _run_gh(["pr", "list", *bb, *hh, "--json",
                   "number,author,headRepositoryOwner,headRepository,headRefName"])
    data = json.loads(out)
    return {int(pr['number']):
            f"{pr['headRepositoryOwner']['login']}:{pr['headRepository']['name']}/{pr['headRefName']}"
            for pr in data}

def gh_pr_merge(pr_id, admin=False, delete_branch=False):
    adm = ["--admin"] if admin else []
    db = ["--delete-branch"] if delete_branch else []
    out = _run_gh(["pr", "merge", str(pr_id), "--merge", *adm, *db])
    if out: print(out)

def gh_release_create(version, title, draft=False):
    dr = ["--draft"] if draft else []
    out = _run_gh(["release", "create", version, *dr, "--generate-notes",
                   "--title", title, "--verify-tag"])
    if out: print(out)


def _run_poetry(cmds):
    assert_poetry_installed()
    cmd = run(["poetry", *cmds], capture_output=True)
    if cmd.returncode == 0:
        return cmd.stdout.decode('UTF-8').strip()
    else:
        raise PoetryError(' '.join(cmds) + ":\n" + cmd.stderr.decode('UTF-8').strip())

def poetry_bump_version(bump):
    out = _run_poetry(["version", bump])
    if out: print(out)

def poetry_get_version():
    return _run_poetry(["version"]).split()[-1]

def poetry_get_expected_version(bump):
    return _run_poetry(["version", bump, "--dry-run"]).split()[-1]

def poetry_publish(build=False):
    pp = ["--build"] if build else []
    out = _run_poetry(["publish", *pp])
    if out: print(out)


class GitError(OSError):
    pass

class GhError(OSError):
    pass

class PoetryError(OSError):
    pass



# copyright ############################### #
# This file is part of the Xaux package.    #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import functools

from .function_tools import count_required_arguments


# TODO: allow_underscore_vars_in_init=False is not very robust. Do we need it for ClassProperty?


# Stub class for docstrings and logical class naming
class Singleton:
    # Stub for naming
    def __new__(cls, *args, **kwargs):
        """The __new__ method is expanded to implement the singleton."""
    def __init__(self, *args, **kwargs):
        """The __init__ method is expanded to implement the singleton."""
    def __str__(self):
        """This a default __str__ method for the singleton."""
    def __repr__(self):
        """This a default __repr__ method for the singleton."""
    def __getattribute__(self, name):
        """The __getattribute__ method is expanded to implement the singleton."""
    def delete(cls):
        """The delete() method removes the singleton and invalidates any existing instances,
        allowing to create a new instance the next time the class is instantiated. This is
        useful for resetting the singleton to its default values.
        """
    def get_self(cls, **kwargs):
        """The get_self(**kwargs) method returns the singleton instance, allowing to pass
        any kwargs to the constructor, even if they are not attributes of the singleton.
        This is useful for kwargs filtering in getters or specific functions.
        """
    def filter_kwargs(cls, **kwargs):
        """The filter_kwargs(**kwargs) method splits the kwargs into non-class kwargs and
        class kwargs. This is useful after a call to `get_self(**kwargs)` to only keep the
        kwargs that are not attributes of the singleton.
        """


def singleton(_cls=None, *, allow_underscore_vars_in_init=True):
    """Singleton decorator.
    This decorator will redefine a class (by letting it inherit from itself and renaming it)
    such that only one instance exists and the same instance is returned every time the
    class is instantiated.
    - Each re-initialisation of the singleton will keep the class attributes, and for this
      reason, the __init__ method should not have any required arguments. This is asserted
      at the class creation time.
    - By default, the singleton allows setting private attributes in the constructor,
      but this can be overridden by setting 'allow_underscore_vars_in_init=False'.
    - This decorator is fully compatible with inheritance, and each child of a singleton
      class will be its own singleton.
    - The singleton can be reset by calling the 'delete()' method of the class, which will
      invalidate any existing instances.
    - The decorator provides a get_self method, which is a class method that is more relaxed
      than the constructor, as it allows passing any kwargs even if they aren't attributes
      for the singleton (these  will then just be ignored). This is useful for kwargs
      filtering in getters or specific functions.
    """
    # Caveat: When we need to get the current class of an instance,
    # we should use type(self) instead of self.__class__, as the latter will get into an
    # infinite loop because of the __getattribute__ method.

    # Internal decorator definition to used without arguments
    def decorator_singleton(cls):
        if _has_singleton_parent(cls):
            # No need to decorate this class, as it will be dealt with by the
            # parent's __init_subclass__.
            return cls

        _check_singleton_compatibility(cls)
        wrap_new, wrap_init, wrap_init_subclass, wrap_getattribute = _get_cls_functions(cls)

        @functools.wraps(cls, updated=())
        class LocalSingleton(cls):
            __original_nonsingleton_class__ = cls

            @functools.wraps(wrap_new)
            def __new__(this_cls, *args, **kwargs):
                # If the singleton instance does not exist, create it
                if '_singleton_instance' not in this_cls.__dict__:
                    try:
                        inst = super().__new__(this_cls, *args, **kwargs)
                    except TypeError:
                        # object._new__ does not accept arguments
                        inst = super().__new__(this_cls)
                    inst._initialised = False
                    inst._valid = True
                    this_cls._singleton_instance = inst
                return this_cls._singleton_instance

            @functools.wraps(wrap_init)
            def __init__(self, *args, **kwargs):
                this_cls = type(self)
                # Validate kwargs
                kwargs.pop('_initialised', None)
                for kk in list(kwargs.keys()) + list(args):
                    if not allow_underscore_vars_in_init and kk.startswith('_'):
                        raise AttributeError(f"Cannot set private attribute {kk} for {this_cls.__name__}! "
                                            + "Use the appropriate setter method instead. However, if you "
                                            + "really want to be able to set this attribute in the "
                                            + "constructor, use 'allow_underscore_vars_in_init=True' "
                                            + "in the singleton decorator.")
                # Initialise the singleton if it has not been initialised yet
                if not self._initialised:
                    # self._initialised = True
                    # Call the init of the original class - hence only once (the first initialisation)
                    super().__init__(*args, **kwargs)
                    self._initialised = True
                # Set the attributes; only attributes defined in the class, custom init, or properties
                # are allowed
                for kk, vv in kwargs.items():
                    if not hasattr(self, kk) and not hasattr(this_cls, kk):
                        raise AttributeError(f"Invalid attribute {kk} for {this_cls.__name__}!")
                    setattr(self, kk, vv)

            @functools.wraps(wrap_init_subclass)
            def __init_subclass__(child_cls, *args, **kwargs):
                # We need to ensure all derived classes behave as singletons as well.
                if '_patched_singleton' not in child_cls.__dict__ \
                or not child_cls._patched_singleton:
                    _check_singleton_compatibility(child_cls)
                    wrap_new, wrap_init, _, _ = _get_cls_functions(child_cls)
                    child_new = child_cls.__dict__.get('__new__')
                    child_init = child_cls.__dict__.get('__init__')
                    # In our redefinitions, whenever we call super we have to call the super
                    # of child_cls, not type(self), otherwise we get into an infinite loop
                    # when the singleton is a parent of the direct parent.
                    @functools.wraps(wrap_new)
                    def this_new(this_cls, *args, **kwargs):
                        if '_singleton_instance' not in this_cls.__dict__:
                            if child_new is not None:
                                try:
                                    inst = child_new(this_cls, *args, **kwargs)
                                except TypeError:
                                    # In python 3.8 and 3.9, __new__ is a staticmethod
                                    inst = child_new.__func__(this_cls, *args, **kwargs)
                            else:
                                inst = super(child_cls, this_cls).__new__(this_cls, *args, **kwargs)
                            if not hasattr(inst, '_singleton_instance'):
                                raise RuntimeError(f"Failed to create the singleton "
                                                 + f"{this_cls.__name__}! Make sure that "
                                                 + "__new__ (if defined) calls the parent "
                                                 + " __new__ method.")
                        return this_cls._singleton_instance
                    child_cls.__new__ = this_new
                    @functools.wraps(wrap_init)
                    def this_init(self, *args, **kwargs):
                        if not self._initialised:
                            if child_init is not None:
                                child_init(self, *args, **kwargs)
                            else:
                                super(child_cls, self).__init__(*args, **kwargs)
                            if not hasattr(self, '_initialised') \
                            or not self._initialised:
                                raise RuntimeError(f"Failed to initialise the singleton "
                                                 + f"{type(self).__name__}! Make sure "
                                                 + "that __init__ (if defined) calls the parent "
                                                 + "parent  __init__ method.")
                        else:
                            super(child_cls, self).__init__(*args, **kwargs)
                    child_cls.__init__ = this_init
                    child_cls._patched_singleton = True
                super().__init_subclass__(*args, **kwargs)

            if cls.__dict__.get('__str__') is None:
                @functools.wraps(Singleton.__str__)
                def __str__(self):
                    return f"<{type(self).__name__} singleton instance>"

            if cls.__dict__.get('__repr__') is None:
                @functools.wraps(Singleton.__repr__)
                def __repr__(self):
                    return f"<{type(self).__name__} singleton instance at {hex(id(self))}>"

            @functools.wraps(wrap_getattribute)
            def __getattribute__(self, name):
                this_cls = type(self)
                if not hasattr(this_cls, '_singleton_instance') \
                or not super().__getattribute__('_valid'):
                    raise RuntimeError(f"This instance of the singleton {this_cls.__name__} "
                                      + "has been invalidated!")
                return super().__getattribute__(name)

            @classmethod
            @functools.wraps(Singleton.delete)
            def delete(this_cls):
                if '_singleton_instance' in this_cls.__dict__:
                    # Invalidate (pointers to) existing instances!
                    this_cls._singleton_instance._valid = False
                    del this_cls._singleton_instance

            @classmethod
            @functools.wraps(Singleton.get_self)
            def get_self(this_cls, **kwargs):
                _, filtered_kwargs = this_cls.filter_kwargs(**kwargs)
                if '_singleton_instance' not in this_cls.__dict__:
                    self = this_cls()
                else:
                    self = this_cls._singleton_instance
                for kk, vv in filtered_kwargs.items():
                    setattr(self, kk, vv)
                return self

            @classmethod
            @functools.wraps(Singleton.filter_kwargs)
            def filter_kwargs(this_cls, **kwargs):
                # Need to initialise in case the instance does not yet exist
                # (to recognise the allowed fields)
                if '_singleton_instance' not in this_cls.__dict__:
                    self = this_cls()
                else:
                    self = this_cls._singleton_instance
                cls_kwargs = {key: value for key, value in kwargs.items()
                                if hasattr(this_cls, key) \
                                or hasattr(this_cls._singleton_instance, key)}
                if not allow_underscore_vars_in_init:
                    cls_kwargs = {key: value for key, value in cls_kwargs.items()
                                    if not key.startswith('_')}
                non_cls_kwargs = kwargs.copy()
                for kk in cls_kwargs.keys():
                    non_cls_kwargs.pop(kk)
                return non_cls_kwargs, cls_kwargs

        # Rename the original class, for clarity in the __mro__ etc
        cls.__name__ = f"{cls.__name__}Original"
        cls.__qualname__ = f"{cls.__qualname__}Original"

        return LocalSingleton

    # Hack to allow the decorator to be used with or without arguments
    if _cls is None:
        return decorator_singleton
    else:
        return decorator_singleton(_cls)


def _has_singleton_parent(cls):
    return any(hasattr(cc, '__original_nonsingleton_class__') for cc in cls.__mro__)


def _check_singleton_compatibility(cls):
    # Verify any existing __init__ method only has optional values
    original_init = cls.__dict__.get('__init__')
    if original_init is not None and count_required_arguments(original_init) > 1:
        raise TypeError(f"Cannot create a singleton for class {cls.__name__} with an "
                        + f"__init__ method that has more than one required argument (only "
                        + f"'self' is allowed)!")

    # Check the class doesn't already have a get_self method
    if cls.__dict__.get('get_self') is not None:
        raise TypeError(f"Class {cls.__name__} provides a 'get_self' method. This is not "
                        + "compatible with the singleton decorator!")

    # Check the class doesn't already have a delete method
    if cls.__dict__.get('delete') is not None:
        raise TypeError(f"Class {cls.__name__} provides a 'delete' method. This is not "
                        + "compatible with the singleton decorator!")


def _get_cls_functions(cls):
        # Define wrapper names
        wrap_new = cls.__new__ if cls.__dict__.get('__new__') is not None \
                               else Singleton.__new__
        wrap_init = cls.__init__ if cls.__dict__.get('__init__') is not None \
                                 else Singleton.__init__
        wrap_init_subclass = cls.__init_subclass__ if cls.__dict__.get('__init_subclass__') is not None \
                                 else Singleton.__init_subclass__
        wrap_getattribute = cls.__getattribute__ if cls.__dict__.get('__getattribute__') \
                                               is not None else Singleton.__getattribute__
        return wrap_new, wrap_init, wrap_init_subclass, wrap_getattribute



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os, sys
from pathlib import Path, PurePosixPath, PureWindowsPath

from .fs import FsPath, _non_strict_resolve
from .eos_methods import _eos_path, _eos_exists, _eos_stat, _eos_lstat, _eos_is_file, \
                         _eos_is_dir, _eos_is_symlink, _eos_touch, _eos_symlink_to, \
                         _eos_unlink, _eos_mkdir, _eos_rmdir, _eos_rmtree, _eos_size

# Note: /eos itself is not on EOS (it is a mountpoint on the local disk)
def _on_eos(*args):
    if isinstance(args[0], str):
        if args[0].startswith('root://eos'):
            return True
        elif args[0].startswith('root:'):
            raise ValueError(f"Unknown EosPath specification {args}.")
    #     # We cannot recognise path file systems by string because of symlinks
    #     elif args[0].startswith('/eos/'):
    #         return True
    #     elif args[0] == '/' and len(args) > 1 \
    #     and (args[1].startswith('eos/') or \
    #         (args[1] == 'eos' and len(args) > 2)):
    #         return True
    absolute = _non_strict_resolve(Path(*args).expanduser().absolute().parent)
    if absolute == _eos_path:
        return True
    parents = list(absolute.parents)
    return len(parents) > 1 and parents[-2] == _eos_path

def _parse_instance(eos_instance):
    if eos_instance == 'home':
        return 'user'
    return eos_instance.lower()

def _parse_mgm(_eos_mgm):
    from xaux.fs.eos_methods import EOS_CELL
    if not isinstance(_eos_mgm, str):
        raise ValueError("The variable `_eos_mgm` should be a string.")
    if not _eos_mgm.startswith('root:'):
        raise ValueError("The variable `_eos_mgm` should start with 'root:'!")
    parts = _eos_mgm.split('/')
    if parts[1] != '' or not parts[2].startswith('eos'):
        raise ValueError(f"Invalid EosPath specification: {_eos_mgm}. "
                        + "The full path should start with 'root://eos...'.")
    if len(parts) >= 3 and parts[3] != '':
        raise ValueError(f"Invalid EosPath specification: {_eos_mgm}. "
                        + "The full path should have a double slash // "
                        + "in between the MGM and the path itself. \n"
                        + "Example: 'root://eosuser.cern.ch//eos/user/s/...'.")
    mgm_parts = parts[2].split('.')
    eos_instance = _parse_instance(mgm_parts[0].replace('eos', ''))
    if '.'.join(mgm_parts[1:]).lower() != EOS_CELL.lower():
        raise ValueError(f"This instance is on the cell {mgm_parts[1:]}, "
                       + f"but the code is configured for cell {EOS_CELL}. "
                       + f"Please set `xaux.fs.EOS_CELL = {mgm_parts[1:]}` "
                       + f"after importing xaux.")
    mgm = f'root://eos{eos_instance}.{EOS_CELL.lower()}'
    eos_path = '/'.join(parts[3:]) if len(parts) >= 3 else None
    return mgm, eos_instance, eos_path


class EosPath(FsPath, Path):
    """Path subclass for EOS paths.

    Instantiating an FsPath should call this class.
    """
    __slots__ = ('mgm', 'eos_instance', 'eos_path', 'eos_path_full', '_init_args')


    @classmethod
    def _new(cls, *args, _eos_checked=False):
        if cls is EosPath:
            cls = EosWindowsPath if os.name == 'nt' else EosPosixPath
        with cls._in_constructor():
            try:
                self = cls._from_parts(args)
            except AttributeError:
                self = Path.__new__(cls, *args)
        with cls._in_constructor(_force=True):
            if not _eos_checked and not _on_eos(*args):
                raise ValueError("The path is not on EOS.")
        return self

    def __new__(cls, *args, _eos_checked=False, _eos_mgm=None):
        if len(args) == 0:
            args = ('.',)
        if isinstance(args[0], str) \
        and args[0].startswith('root:'):
            if len(args) > 1:
                raise ValueError("When specifying the instance `root://eos...` "
                               + "in the path, the latter has to be given as "
                               + "one complete string.")
            mgm, eos_instance, eos_path = _parse_mgm(args[0])
            if eos_path is None:
                raise ValueError("Incomplete MGM path (only MGM instance was given).")
            self = cls._new(eos_path, _eos_checked=_eos_checked)
            self.eos_instance = eos_instance # Temporary (verified in _set_eos_path in init)
            if sys.version_info >= (3, 12):
                self._init_args = [eos_path]
            if _eos_mgm is not None and mgm != _parse_mgm(_eos_mgm)[0]:
                raise ValueError(f"The specified MGM {_eos_mgm} does not match "
                               + f"the resolved one {mgm}.")
            self.mgm = mgm
        else:
            self = cls._new(*args, _eos_checked=_eos_checked)
            if sys.version_info >= (3, 12):
                self._init_args = args
            if _eos_mgm is not None:
                mgm, eos_instance, _ = _parse_mgm(_eos_mgm)
                self.eos_instance = eos_instance # Temporary (verified in _set_eos_path in init)
                self.mgm = mgm
            else:
                self.eos_instance = None
                self.mgm = None
        return self

    def __init__(self, *args):
        with self.__class__._in_constructor():
            if sys.version_info >= (3, 12):
                Path.__init__(self, *self._init_args)
            else:
                Path.__init__(self)
        if self.eos_instance is None:
            self._set_eos_path()
        else:
            self._set_eos_path(_eos_instance=self.eos_instance)
        if self.mgm is None:
            from xaux.fs.eos_methods import EOS_CELL
            self.mgm = f'root://eos{self.eos_instance}.{EOS_CELL.lower()}'
        self.eos_path_full = f'{self.mgm}/{self.eos_path}'

    def _set_eos_path(self, _eos_instance=None):
        with self.__class__._in_constructor(_force=True):
            parts = _non_strict_resolve(self.expanduser().parent, _as_posix=True).split('/')
            if len(parts) == 2:
                parts = _non_strict_resolve(self.expanduser(), _as_posix=True).split('/')
        instance_parts = parts[2].split('-')
        eos_instance = _parse_instance(instance_parts[0])
        if _eos_instance is not None and eos_instance != _eos_instance:
            raise ValueError(f"This path is on the EOS instance {eos_instance}, "
                           + f"however, the MGM specified it as {_eos_instance}.")
        if len(instance_parts) > 1:
            if len(instance_parts) > 2:
                raise ValueError(f"EOS instance {parts[2]} has more than one dash.")
            else:
                if len(instance_parts[1]) > 1:
                    # This happens e.g. on SWAN where home is cast to home-i04 etc.
                    # We then just completely ignore the i04 part.
                    parts = ['', 'eos', eos_instance, *parts[3:]]
                else:
                    parts = ['', 'eos', eos_instance, instance_parts[1], *parts[3:]]
        else:
            parts = ['', 'eos', eos_instance, *parts[3:]]
        self.eos_path = '/'.join(parts) + '/' + self.name
        self.eos_instance = eos_instance


    # Overwrite Path methods
    # ======================
    if sys.version_info >= (3, 12):
        def exists(self, *args, follow_symlinks=True, **kwargs):
            if self.is_symlink() and follow_symlinks:
                return self.resolve().exists(*args, **kwargs)
            return _eos_exists(self.expanduser(), *args, **kwargs)

    else:
        def exists(self, *args, **kwargs):
            if self.is_symlink():
                return self.resolve().exists(*args, **kwargs)
            return _eos_exists(self.expanduser(), *args, **kwargs)

    if sys.version_info >= (3, 13):
        def is_file(self, *args, follow_symlinks=True, **kwargs):
            if self.is_symlink() and follow_symlinks:
                return self.resolve().is_file(*args, **kwargs)
            return _eos_is_file(self.expanduser(), *args, **kwargs)

        def is_dir(self, *args, follow_symlinks=True, **kwargs):
            if self.is_symlink() and follow_symlinks:
                return self.resolve().is_dir(*args, **kwargs)
            return _eos_is_dir(self.expanduser(), *args, **kwargs)

    else:
        def is_file(self, *args, **kwargs):
            if self.is_symlink():
                return self.resolve().is_file(*args, **kwargs)
            return _eos_is_file(self.expanduser(), *args, **kwargs)

        def is_dir(self, *args, **kwargs):
            if self.is_symlink():
                return self.resolve().is_dir(*args, **kwargs)
            return _eos_is_dir(self.expanduser(), *args, **kwargs)

    def is_symlink(self, *args, **kwargs):
        return _eos_is_symlink(self.expanduser(), *args, **kwargs)

    def stat(self, *args, **kwargs):
        return _eos_stat(self.expanduser(), *args, **kwargs)

    def lstat(self, *args, **kwargs):
        return _eos_lstat(self.expanduser(), *args, **kwargs)

    def touch(self, *args, **kwargs):
        return _eos_touch(self.expanduser(), *args, **kwargs)

    def unlink(self, *args, **kwargs):
        return _eos_unlink(self.expanduser(), *args, **kwargs)

    def mkdir(self, *args, **kwargs):
        return _eos_mkdir(self.expanduser(), *args, **kwargs)

    def rmdir(self, *args, **kwargs):
        return _eos_rmdir(self.expanduser(), *args, **kwargs)

    def as_posix(self, *args, **kwargs):
        if hasattr(self, 'eos_path'):
            return self.eos_path
        return Path.as_posix(self, *args, **kwargs)

    # def glob(self, *args, **kwargs):
    #     raise NotImplementedError

    # def rglob(self, *args, **kwargs):
    #     raise NotImplementedError


    # Overwrite FsPath methods
    # ======================

    def symlink_to(self, target, target_is_directory=False, **kwargs):
        target = FsPath(target)
        return _eos_symlink_to(self.expanduser(), FsPath, target.expanduser(), target_is_directory=target_is_directory, **kwargs)

    def rmtree(self, *args, **kwargs):
        return _eos_rmtree(self.expanduser(), FsPath, *args, **kwargs)

    def size(self, *args, **kwargs):
        return _eos_size(self.expanduser(), FsPath, *args, **kwargs)


class EosPosixPath(EosPath, PurePosixPath):
    """EosPath subclass for EOS paths on non-Windows systems.

    On a POSIX system, instantiating an FsPath or an EosPath
    should return this object.
    """
    __slots__ = ()

    if os.name == 'nt':
        def __new__(cls, *args, **kwargs):
            raise OSError(
                f"Cannot instantiate {cls.__name__!r} on your system")


class EosWindowsPath(EosPath, PureWindowsPath):
    """EosPath subclass for EOS paths on Windows systems.

    This is currently not supported.
    """
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        raise NotImplementedError(
            "Windows currently not supported by EosPath.")

    if os.name != 'nt':
        def __new__(cls, *args, **kwargs):
            raise OSError(
                f"Cannot instantiate {cls.__name__!r} on your system")




# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os, sys
from subprocess import run, PIPE, CalledProcessError
from pathlib import Path, PurePosixPath, PureWindowsPath

from .fs import FsPath, _non_strict_resolve
from .fs_methods import _xrdcp_installed


_afs_path = Path('/afs')
_afs_mounted = False
try:
    cmd = run(['mount'], stdout=PIPE, stderr=PIPE)
    if cmd.returncode == 0:
        stdout = cmd.stdout.decode('UTF-8').strip()
        mounts = [line.split(' on ')[1].split()[0] for line in stdout if 'AFS' in line]
        if len(mounts) == 1:
            _afs_mounted = True
            _afs_path = Path(mounts[0])
        elif len(mounts) > 1:
            raise OSError("Multiple AFS mounts detected.")
except (CalledProcessError, FileNotFoundError):
    pass

try:
    cmd = run(['fs', '--version'], stdout=PIPE, stderr=PIPE)
    _fs_installed =  cmd.returncode == 0
except (CalledProcessError, FileNotFoundError):
    _fs_installed = False

def _assert_fs_installed(mess=None):
    if not _fs_installed:
        mess = f" {mess}" if mess is not None else mess
        raise OSError(f"`fs` is not installed on your system.{mess}")

_afs_mounted = _afs_path.exists()
afs_accessible = _afs_mounted or _xrdcp_installed

def _assert_afs_accessible(mess=None):
    if not afs_accessible:
        mess = f" {mess}" if mess is not None else mess
        raise OSError(f"AFS is not installed on your system.{mess}")

# Note: /afs itself is not on AFS (it is a mountpoint on the local disk)
def _on_afs(*args):
    # We cannot recognise path file systems by string because of symlinks
    # if isinstance(args[0], str):
    #     if args[0].startswith('/afs/'):
    #         return True
    #     elif args[0] == '/' and len(args) > 1 \
    #     and (args[1].startswith('afs/') or \
    #         (args[1] == 'afs' and len(args) > 2)):
    #         return True
    absolute = _non_strict_resolve(Path(*args).expanduser().absolute().parent)
    if absolute == _afs_path:
        return True
    parents = list(absolute.parents)
    return len(parents) > 1 and parents[-2] == _afs_path


class AfsPath(FsPath, Path):
    """Path subclass for AFS paths.

    Instantiating an FsPath should call this class.
    """
    __slots__ = ('afs_cell')

    def __new__(cls, *args, _afs_checked=False):
        if cls is AfsPath:
            cls = AfsWindowsPath if os.name == 'nt' else AfsPosixPath
        with cls._in_constructor():
            try:
                self = cls._from_parts(args)
            except AttributeError:
                self = Path.__new__(cls, *args)
        with cls._in_constructor(_force=True):
            if not _afs_checked and not _on_afs(*args):
                raise ValueError("The path is not on AFS.")
        return self

    def __init__(self, *args):
        with self.__class__._in_constructor():
            if sys.version_info >= (3, 12):
                Path.__init__(self, *args)
            else:
                Path.__init__(self)
        with self.__class__._in_constructor(_force=True):
            parent = self.parent
        if parent == _afs_path:
            self.afs_cell = self.name
        else:
            res_parts = _non_strict_resolve(parent, _as_posix=True).split('/')
            if len(res_parts) < 3:
                raise ValueError("Malformed AfsPath.")
            self.afs_cell = res_parts[2].upper()

    # Overwrite Path methods
    # ======================

    def exists(self, *args, **kwargs):
        _assert_afs_accessible("Cannot check for existence of AFS paths.")
        return Path.exists(self.expanduser(), *args, **kwargs)

    def touch(self, *args, **kwargs):
        _assert_afs_accessible("Cannot touch AFS paths.")
        return Path.touch(self, *args, **kwargs)

    def symlink_to(self, *args, **kwargs):
        _assert_afs_accessible("Cannot create symlinks on AFS paths.")
        return Path.symlink_to(self.expanduser(), *args, **kwargs)

    def getfid(self):
        if _fs_installed:
            cmd = run(['fs', 'getfid', self.expanduser().as_posix()],
                      stdout=PIPE, stderr=PIPE)
            if cmd.returncode == 0:
                stdout = cmd.stdout.decode('UTF-8').strip()
                if "File" in stdout:
                    return stdout.split()[2][1:-1]
        FsPath.getfid(self)

    def flush(self):
        if _fs_installed:
            cmd = run(['fs', 'flush', self.expanduser().as_posix()],
                      stdout=PIPE, stderr=PIPE)
            if cmd.returncode == 0:
                self.touch()
                return
        FsPath.flush(self)

    # New methods
    # ======================

    @property
    def acl(self):
        _assert_fs_installed("Cannot get ACL of AFS paths.")
        cmd = run(['fs', 'la', self.expanduser().as_posix()],
                            stdout=PIPE, stderr=PIPE)
        if cmd.returncode == 0:
            acl = {}
            output = cmd.stdout.decode('UTF-8').strip()
            lines = output.split('\n')
            if FsPath(lines[0].split()[-2]).resolve() != self.resolve():
                raise OSError(f"Cannot interpret ACL output.\n{output}")
            for line in lines[1:]:
                if line.startswith('  '):
                    parts = line.split()
                    acl[parts[0]] = parts[1]
                    if len(parts) > 2:
                        raise OSError(f"Cannot interpret ACL output.\n{output}")
                elif line != 'Normal rights:':
                    raise OSError(f"Cannot interpret ACL output.\n{output}")
            if acl == {}:
                raise OSError(f"Cannot interpret ACL output.\n{output}")
            return acl
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            raise OSError(f"Failed to retrieve ACL on {self}.\n{stderr}")

    @acl.setter
    def acl(self, val):
        _assert_fs_installed("Cannot set ACL of AFS paths.")
        if not isinstance(val, dict):
            raise ValueError("ACL has to be a dictionary of users and permissions.")
        for user, acl in val.items():
            if acl is None or acl == '':
                acl = 'none'
            if not isinstance(user, str):
                raise ValueError("User in ACL has to be a string.")
            if not isinstance(acl, str):
                raise ValueError("ACL has to be a string or `None`.")
            cmd = run(['fs', 'sa', self.expanduser().as_posix(), user, acl],
                                stdout=PIPE, stderr=PIPE)
            if cmd.returncode != 0:
                stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
                raise OSError(f"Failed to set ACL to {acl} on {self} for user "
                                 + f"{user}.\n{stderr}")

    @acl.deleter
    def acl(self):
        _assert_fs_installed("Cannot delete ACL of AFS paths.")
        cmd = run(['fs', 'sa', self.expanduser().as_posix(), 'none'],
                            stdout=PIPE, stderr=PIPE)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            raise OSError(f"Failed to delete ACL on {self}.\n{stderr}")


class AfsPosixPath(AfsPath, PurePosixPath):
    """AfsPath subclass for AFS paths on non-Windows systems.

    On a POSIX system, instantiating an FsPath or an AfsPath
    should return this object.
    """
    __slots__ = ()

    if os.name == 'nt':
        def __new__(cls, *args, **kwargs):
            raise OSError(
                f"Cannot instantiate {cls.__name__!r} on your system")


class AfsWindowsPath(AfsPath, PureWindowsPath):
    """AfsPath subclass for AFS paths on Windows systems.

    On a Windows system, instantiating an FsPath or an AfsPath
    should return this object.
    """
    __slots__ = ()

    if os.name != 'nt':
        def __new__(cls, *args, **kwargs):
            raise OSError(
                f"Cannot instantiate {cls.__name__!r} on your system")





# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

# Last update 28/10/2024 - T. Pugnat and F.F. Van der Veken

import os
import io
import sys
import time
import json
import atexit
import signal
import random
import traceback

from ..fs import FsPath, EosPath
from ..fs.temp import _tempdir
from .general_tools import ranID, get_hash, timestamp


protected_open = {}


# The functions registered via this module are not called when the program is killed by a signal not handled by Python, when a Python fatal internal error is detected, or when os._exit() is called.
def exit_handler():
    """This handles cleaning of potential leftover lockfiles."""
    for file in protected_open.values():
        file.release(pop=False)

# This one should handle those exceptions.
def kill_handler(signum, frame):
    exit_handler()
    print(f"\n\nTraceback (most recent call last):")
    traceback.print_stack(frame)
    print(f"{signal.Signals(signum).name}: [Errno {signum}] A signal has been raised.")
    sys.exit(0)

def _register_exithandlers(obj):
    if not hasattr(obj.__class__, '_exithandler_registered') \
    or not obj.__class__._exithandler_registered:
        atexit.register(exit_handler)
        signal.signal(signal.SIGINT, kill_handler)
        signal.signal(signal.SIGTERM, kill_handler)
        obj.__class__._exithandler_registered = True


# TODO: there is some issue with the timestamps. Was this really a file
#       corruption, or is this an OS issue that we don't care about?
# TODO: no stats on EOS files
def get_fstat(filename):
    stats = FsPath(filename).stat()
    return {
                'n_sequence_fields': int(stats.n_sequence_fields),
                'n_unnamed_fields':  int(stats.n_unnamed_fields),
                'st_mode':           int(stats.st_mode),
                'st_ino':            int(stats.st_ino),
                'st_dev':            int(stats.st_dev),
                'st_uid':            int(stats.st_uid),
                'st_gid':            int(stats.st_gid),
                'st_size':           int(stats.st_size),
                'st_mtime_ns':       int(stats.st_mtime_ns),
                'st_ctime_ns':       int(stats.st_ctime_ns),
            }


class ProtectFile:
    """A wrapper around a file pointer, protecting it with a lockfile.

    Use
    ---
    It is meant to be used inside a context, where the entering and leaving of a
    context ensures file protection. The moment the object is instantiated, a
    lockfile is generated (which is destroyed after leaving the context). Attempts
    to access the file will be postponed as long as a lockfile exists. Furthermore,
    while in the context, file operations are done on a temporary file, that is
    only moved back when leaving the context.

    The reason to lock read access as well is that we might work with immutable
    files. The following scenario might happen: a file is read by process 1, some
    calculations are done by process 1, the file is read by process 2, and the
    result of the calculations is written by process 1. Now process 2 is working
    on an outdated version of the file. Hence the full process should be locked in
    one go: reading, manipulating/calculating, and writing.

    An important caveat is that, after the manipulation/calculation, the file
    contents have to be wiped before writing, otherwise the contents will be
    appended (as the file pointer is still at the end of the file after reading it
    in). Unless of course that is the intended result. Wiping the file can be
    achieved with the built-in truncate() and seek() methods.

    Attributes
    ----------
    file       : pathlib.Path
        The path to the file to be protected.
    lockfile   : pathlib.Path
        The path to the lockfile.
    tempfile   : pathlib.Path
        The path to a temporary file which will accumulate all writes until the
        ProtectFile object is destroyed, at which point the temporary file will
        replace the original file. Not used when a ProtectFile object is
        instantiated in read-only mode ('r' or 'rb').

    Examples
    --------
    Reading in a file (while making sure it is not written to by another process):

    >>> from xaux import ProtectFile
    >>> with ProtectFile('thebook.txt', 'r', wait=1) as pf:
    >>>    text = pf.read()

    Reading and appending to a file:

    >>> from xaux import ProtectFile
    >>> with ProtectFile('thebook.txt', 'r+', wait=1) as pf:
    >>>    text = pf.read()
    >>>    pf.write("This string will be added at the end of the file, \
    ...               however, it won't be added to the 'text' variable")

    Reading and updating a JSON file:

    >>> import json
    >>> from xaux import ProtectFile
    >>> with ProtectFile(info.json, 'r+', wait=1) as pf:
    >>>     meta = json.load(pf)
    >>>     meta.update({'author': 'Emperor Claudius'})
    >>>     pf.truncate(0)          # Delete file contents (to avoid appending)
    >>>     pf.seek(0)              # Move file pointer to start of file
    >>>     json.dump(meta, pf, indent=2, sort_keys=False))

    Reading and updating a Parquet file:

    >>> import pandas as pd
    >>> from xaux import ProtectFile
    >>> with ProtectFile(mydata.parquet, 'r+b', wait=1) as pf:
    >>>     data = pd.read_parquet(pf)
    >>>     data['x'] += 5
    >>>     pf.truncate(0)          # Delete file contents (to avoid appending)
    >>>     pf.seek(0)              # Move file pointer to start of file
    >>>     data.to_parquet(pf, index=True)
    """

    # Use debug flag below to inspect steps in file IO
    _debug = False
    _testing = False


    def __init__(self, *args, **kwargs):
        """A ProtectFile object, to be used only in a context.

        Parameters
        ---------
        wait : float, default 1
            When a file is locked, the time to wait in seconds before trying to
            access it again.
        use_temporary : bool, default True
            Whether or not to perform writing operations on a temporary file.
            Ignored when the file is read-only.
        check_hash : bool, default True
            Whether or not to verify by hash that the file did not change during
            the lock.
        max_lock_time : float, default None
            If provided, it will write the maximum runtime in seconds inside the
            lockfile. This is to avoid crashed accesses locking the file forever.

        Additionally, the following parameters are inherited from open():
            'file', 'mode', 'buffering', 'encoding', 'errors', 'newline', 'closefd', 'opener'
        """
        _register_exithandlers(self)

        # File variables
        # ==============
        # self._file:   path to the file to the protected file
        # self._lock:   path to the lockfile
        # self._temp:   path to the temporary file on which write operations are applied
        # self._fd:     file pointer to the main file
        #               (self._file if readonly, self._temp if writing)

        argnames_open = ['file', 'mode', 'buffering', 'encoding', 'errors', 'newline',
                         'closefd', 'opener']
        arg = dict(zip(argnames_open, args))
        arg.update(kwargs)

        # Using a temporary file to write to
        self._use_temporary = arg.pop('use_temporary', True)
        self._check_hash = arg.pop('check_hash', True)

        # Initialise paths
        arg['file'] = FsPath(arg['file']).resolve()
        file = arg['file']
        self._file = file
        self._lock = FsPath(file.parent, file.name + '.lock').resolve()
        self._temp = FsPath(_tempdir.name, file.name + ranID()).resolve()

        # We throw potential FileNotFoundError and FileExistsError before
        # creating the temporary file
        self._exists = True if self.file.is_file() else False
        if not self._exists and self.file.exists():
            raise NotImplementedError("ProtectFile does not yet support "
                                    + "directories or symlinks.")
        mode = arg.get('mode','r')
        self._readonly = False
        if 'r' in mode:
            if not self._exists:
                raise FileNotFoundError
            if not '+' in mode:
                self._readonly = True
        elif 'x' in mode:
            if self._exists:
                raise FileExistsError
        if self._readonly:
            self._use_temporary = False

        # Provide an expected running time (to free a file in case of crash)
        max_lock_time = arg.pop('max_lock_time', None)
        if max_lock_time is not None and self._readonly == False:
            print("Warning: Using `max_lock_time` for non read-only "
                + "files is dangerous! If the time is estimated wrongly "
                + "and a file is freed while the original process is "
                + "still running, file corruption might occur. Are you "
                + "sure this is what you want?")
        if max_lock_time and max_lock_time < 2:
            print("Warning: `max_lock_time` is too short. Put to 2.")
            max_lock_time = 2

        # Time to wait between trials to generate lockfile
        wait = arg.pop('wait', 1)

        # Try to make lockfile, wait if unsuccesful
        self._access = False
        while True:
            try:
                self._create_lock(max_lock_time=max_lock_time)
                # Success! Or is it....?
                # We are in the lockfile, but there is still one potential concurrency,
                # namely another process could have started creating the file while we
                # did not see it having been created yet...
                self._flush_lock(wait=1e-3*wait)
                if not self._lock_is_ours():
                    self._wait(wait)
                    continue
                self._print_debug("init", f"created {self.lockfile}")
                break

            except FileNotFoundError:
                # Lockfile could not be created, wait and try again
                self._wait(wait)
                continue

            except PermissionError:
                # Special case: we can still access eos files when permission has expired, using `eos`
                if isinstance(self.file, EosPath):
                    try:
                        # If it already exists, we have to wait anyway for it to be freed
                        if self.lockfile.is_file():
                            self._wait(wait)
                            continue
                        # Make a local lockfile that has the sysinfo
                        local_lockfile = FsPath(_tempdir.name, file.name + '.lock').resolve()
                        if local_lockfile.exists():
                            local_lockfile.unlink()
                        self._create_lock(local_lockfile, max_lock_time, local=True)
                        self._print_debug("init", f"created local {local_lockfile}")
                        self._flush_lock(local_lockfile, wait=1e-3*wait)
                        if not self._lock_is_ours(local_lockfile):
                            self._wait(wait)
                            continue
                        self._print_debug("init", f"created {self.lockfile} via eos cp")
                        break
                    except PermissionError:
                        # This means the `eos` command has failed as well: we really don't have access
                        raise PermissionError(f"Cannot access {self.lockfile}; permission denied.")
                else:
                    raise PermissionError(f"Cannot access {self.lockfile}; permission denied.")

            except OSError:
                # Two typical cases: the lockfile already exists (FileExistsError, a subclass of OSError),
                # or an input/output error happened while trying to generate it (generic OSError).
                # In both cases, we wait a bit and try again.
                self._wait(wait)
                # We also have to capture the case where the lockfile expired and can be freed.
                # So we try to read it and look for the timeout period; if this fails (e.g. because the
                # lock disappeared in the meanwhile), we continue the mainloop
                try:
                    kill_lock = False
                    with self.lockfile.open('r') as fid:
                        info = json.load(fid)
                    if 'free_after' in info and int(info['free_after']) > 0 \
                    and int(info['free_after']) < time.time():
                        # We free the original process by deleting the lockfile
                        # and then we go to the next step in the while loop.
                        # Note that this does not necessarily imply this process
                        # gets to use the file; which is the intended behaviour
                        # (first one wins).
                        kill_lock = True
                    if kill_lock:
                        self.lockfile.unlink()
                        self._print_debug("init",f"freed {self.lockfile} because "
                                            + "of exceeding max_lock_time")
                    # Whether or not the lockfile was freed, we continue to the main loop
                    continue

                except (OSError, json.JSONDecodeError):
                    # Any error in trying to read (and potentially kill the lock) implies
                    # a return to the main loop
                    continue

        # Success!
        self._access = True
        self._delete_lock_at_finish = True

        # Store stats (to check if file got corrupted later)
        if self._check_hash and self._exists:
            self._size = self.file.size()
            self._hash = get_hash(self.file)

        # Force an update from the file system (a bit slow ~100ms, but necessary)
        self._file.flush()

        # Choose file pointer:
        # To the temporary file if writing, or existing file if read-only
        if self._use_temporary:
            if self._exists:
                self._print_debug("init", f"cp {self.file=} to {self.tempfile=}")
                self.file.copy_to(self.tempfile)
            arg['file'] = self.tempfile
        self._fd = io.open(**arg)

        # Store object in class dict for cleanup in case of sysexit
        protected_open[self.file] = self


    def _wait(self, wait):
        # Add some white noise to the wait time to avoid different processes syncing
        if self._testing:
            this_wait = random.uniform(wait*0.999, wait*1.001)
        else:
            this_wait = random.uniform(wait*0.6, wait*1.4)
        self._print_debug("init", f"waiting {this_wait}s to create {self.lockfile}")
        time.sleep(this_wait)


    def _create_lock(self, lockfile=None, max_lock_time=None):
        self.lockfile.getfid() # Look up the file on the server (takes a few ms)
        if lockfile is None:
            lockfile = self.lockfile
        free_after = -1
        if max_lock_time is not None:
            free_after = int(time.time() + max_lock_time)
        free_after = f"{free_after:15d}"[:15]
        # We ensure that the variables in the lockfile always have a fixed number of characters
        ran = random.randint(0, 2**63 - 1) + os.getpid() + int(time.time_ns() % 1e9)
        self._ran = f"{ran:0>20d}"
        self._machine = f"{os.uname().nodename: >35s}"[:35]
        with lockfile.open('x') as flock:
            json.dump({
                'ran':     self._ran,
                'machine': self._machine,
                'free_after': free_after
            }, flock)
        self._print_debug("init", f"Trying lockfile with metadata {free_after=} ran={self._ran} machine={self._machine}")


    def _flush_lock(self, local_lockfile=None, wait=0.01):
        if local_lockfile:
            # Move it to the server lockfile
            local_lockfile.move_to(self.lockfile)  # can use specialised server commands
            assert not local_lockfile.is_file()    # sanity check
        # Flush the file on the server (a bit slow ~100ms, but necessary)
        self.lockfile.flush()
        if self._testing:
            this_wait = random.uniform(0.099, 0.101)
        else:
            this_wait = 0.001 + random.uniform(wait*0.6, wait*1.4)
        self._print_debug("init", f"flushing lock and waiting {this_wait}s to ensure sync")
        time.sleep(this_wait)
        if local_lockfile:
            # Move it to the server lockfile
            self.lockfile.copy_to(local_lockfile)  # can use specialised server commands
            assert local_lockfile.is_file()    # sanity check


    def _lock_is_empty(self, i=1):
        if i > 15:
            return True
        if self.lockfile.size() > 0:
            return False
        else:
            self._wait(0.2)
            return self._lock_is_empty(i+1)


    def _lock_is_ours(self, lockfile=None):
        if lockfile is None:
            lockfile = self.lockfile
        if not lockfile.exists():
            return False
        if self._lock_is_empty():
            self._print_debug("lock_is_ours", f"lockfile {lockfile} is empty")
            lockfile.unlink()
            return False
        try:
            with lockfile.open('r') as fid:
                info = json.load(fid)
        except:
            # If we cannot load the json, it might be empty or being written to
            self._print_debug("lock_is_ours", f"cannot load json info from {lockfile}")
            return False
        if 'ran' not in info or info['ran'] != self._ran:
            ran = info['ran'] if 'ran' in info else 'None'
            self._print_debug("lock_is_ours", f"ran info changed in {lockfile} ({ran} vs {self._ran})")
            return False
        if 'machine' not in info or info['machine'] != self._machine:
            machine = info['machine'] if 'machine' in info else 'None'
            self._print_debug("lock_is_ours", f"machine info changed in {lockfile} ({machine} vs {self._machine})")
            return False
        # We got here, so the lockfile is ours
        return True


    def __del__(self, *args, **kwargs):
        self.release()

    def __enter__(self, *args, **kwargs):
        return self._fd

    def __exit__(self, *args, **kwargs):
        if not self._access:
            return
        # Close file pointer
        if not self._fd.closed:
            self._fd.close()
        # Check that the lock is still ours
        if not self._lock_is_ours(self.lockfile):
            self._delete_lock_at_finish = False
            self.stop_with_error(f"Lockfile {self.lockfile} is not ours anymore.")
            return
        # Check that we did not run out of time
        try:
            with self.lockfile.open('r') as fid:
                info = json.load(fid)
        except:
            self._delete_lock_at_finish = False
            self.stop_with_error("Loading JSON from lockfile failed.")
            return
        if 'free_after' in info and int(info['free_after']) > 0 and int(info['free_after']) < time.time():
            # Max runtime was expired. We have to forfeit the job as this is a potential failure point.
            self.stop_with_error(f"Error: Job {self._file} took longer than expected ("
                + f"{round(time.time() - int(info['free_after']))}s. Increase max_lock_time.")
            return
        # Check that original file was not modified in between (i.e. corrupted)
        file_changed = False
        if self._use_temporary and self._check_hash and self._exists:
            new_size = self.file.size()
            new_hash = get_hash(self.file)
            if self._hash != new_hash:
                file_changed = True
            # for key, val in self._fstat.items():
            #     if key not in new_stats or val != new_stats[key]:
            #         file_changed = True
        if file_changed:
            self.stop_with_error(f"Error: File {self.file} changed during lock! "
                + f"Original size: {self._size}, new size: {new_size}. "
                + f"Original hash: {self._hash}, new hash: {new_hash}.")
        else:
            # All is fine: move result from temporary file to original
            self.mv_temp()
            # Flag the changes to the server
            self.file.getfid()
            time.sleep(random.uniform(0.1, 0.2))
            self.release()


    @property
    def file(self):
        return self._file

    @property
    def lockfile(self):
        return self._lock

    @property
    def tempfile(self):
        return self._temp


    def mv_temp(self, destination=None):
        """Move temporary file to 'destination' (the original file if destination=None)"""
        if not self._access:
            return
        if self._use_temporary:
            if destination is None:
                # Move temporary file to original file
                self._print_debug("mv_temp", f"cp {self.tempfile=} to {self.file=}")
                self.tempfile.copy_to(self.file)
                # # Check if copy succeeded
                # if self._check_hash and get_hash(self.tempfile) != get_hash(self.file):
                #     self.stop_with_error(f"Warning: tried to copy temporary file {self.tempfile} into {self.file}, "
                #           + "but hashes do not match!")
            else:
                self._print_debug("mv_temp", f"cp {self.tempfile=} to {destination=}")
                self.tempfile.copy_to(destination)
            self._print_debug("mv_temp", f"unlink {self.tempfile=}")
            self.tempfile.unlink()


    def stop_with_error(self, message):
        """Fail the job, and potentially save calculation results"""
        if not self._access:
            return
        self._access = False
        results_saved = False
        alt_file = None
        if self._use_temporary:
            extension = f"__{timestamp(us=True, in_filename=True)}.result"
            alt_file = FsPath(self.file.parent, self.file.name + extension).resolve()
            self.mv_temp(alt_file)
            results_saved = True
        raise ProtectFileError(message, results_saved, alt_file)


    def release(self, pop=True):
        """Clean up lockfile and tempfile"""
        # Overly verbose in checking, as to make sure this never fails
        # (to avoid being stuck with remnant lockfiles)
        # Close main file pointer
        if hasattr(self,'_fd') and hasattr(self._fd,'closed') and not self._fd.closed:
            self._fd.close()
        # Delete temporary file
        if hasattr(self,'_temp') and hasattr(self._temp,'is_file') and self._temp.is_file():
            self._print_debug("release", f"unlink {self.tempfile}")
            self.tempfile.unlink()
        # Delete lockfile
        if hasattr(self, '_delete_lock_at_finish') and self._delete_lock_at_finish:
            if hasattr(self,'_lock') and hasattr(self._lock,'is_file') and self._lock.is_file():
                self._print_debug("release", f"unlink {self.lockfile}")
                self.lockfile.unlink()
        # Remove file from the protected register
        if pop and hasattr(self, '_file'):
            protected_open.pop(self._file, 0)


    def _print_debug(self, prc, msg):
        if self._debug:
            print(f"({self._file.name}) {prc}: {msg}")


class ProtectFileError(Exception):
    def __init__(self, message, results_saved, alt_file):
        mess = f"ProtectFile failed! {message}"
        if results_saved:
            mess += f" Saved calculation results in {alt_file.name}."
        super().__init__(mess)
        self.message = message
        self.results_saved = results_saved
        self.alt_file = alt_file

    def __reduce__(self):
        return (ProtectFileError, (self.message, self.results_saved, self.alt_file))



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import sys
import json
import importlib
import urllib.request
from shutil import rmtree
from subprocess import run
from contextlib import contextmanager
from packaging.version import Version

from ..general import _pkg_root


_PACKAGE_PATH = _pkg_root / "lib" / "_xaux_pkg_vers"

xsuite_pkgs = ['xaux', 'xobjects', 'xdeps', 'xtrack', 'xpart', 'xfields', 'xcoll', 'xdyna', 'xboinc']


@contextmanager
def import_package_version(package_name, version, wipe_cache=False):
    """
    Context manager to temporarily import a specific version of a package.
    """
    package_path = _PACKAGE_PATH / package_name / version
    if not package_path.exists():
        install_package_version(package_name, version)

    original_sys_path = sys.path.copy()
    original_sys_modules = sys.modules.copy()
    original_sys_meta_path = sys.meta_path.copy()
    pkgs = _get_available_packages_in_path(package_path)
    for pkg in pkgs:
        _remove_import_from_sys(pkg)
    for finder_name in sys.meta_path:
        if package_name in str(finder_name):
            sys.meta_path.remove(finder_name)
    if wipe_cache:
        importlib.invalidate_caches()
        for finder_name in sys.meta_path:
            if 'editable' in str(finder_name):
                sys.meta_path.remove(finder_name)
    sys.path.insert(0, package_path.as_posix())
    try:
        module = importlib.import_module(package_name)
        yield module
    finally:
        sys.path = original_sys_path
        sys.modules = original_sys_modules
        sys.meta_path = original_sys_meta_path


def install_package_version(package_name, version, overwrite=False):
    """
    Installs a specific version of a package.
    """
    full_package_name = f"{package_name}=={version}"
    package_path = _PACKAGE_PATH / package_name / version
    if package_path.exists() and not overwrite:
        print(f"Package already installed in {package_path}. "
             + "Use `overwrite=True` if this is the intended goal.")
    else:
        if package_path.exists():
            rmtree(package_path)
        package_path.mkdir(parents=True)
        try:
            print(f"Installing package {full_package_name}")
            cmd = run([sys.executable, "-m", "pip", "install", full_package_name,
                    f"--target={package_path}"], capture_output=True)
            if cmd.returncode != 0:
                print(f"Installation failed for {full_package_name}:\n{cmd.stderr.decode()}")
        except Exception as e:
            print(f"An error occurred: {e}\nCommand output:{cmd.stderr.decode()}")


def get_package_versions(package_name):
    """
    Get all available versions of a package from PyPI, sorted by newest last.
    """
    data = json.loads(urllib.request.urlopen(f"https://pypi.org/pypi/{package_name}/json").read())
    return sorted(list(data['releases'].keys()), key=Version)


def get_latest_package_version(package_name):
    """
    Get the latest version of a package from PyPI.
    """
    data = json.loads(urllib.request.urlopen(f"https://pypi.org/pypi/{package_name}/json").read())
    return data['info']['version']


def get_package_dependencies(package_name):
    """
    Get the dependencies of a package from PyPI.
    """
    data = json.loads(urllib.request.urlopen(f"https://pypi.org/pypi/{package_name}/json").read())
    return data['info']['requires_dist']


def get_package_version_dependencies(package_name, version, skip=[]):
    """
    Get the package versions of the dependencies for a specific version of `package_name`.
    TODO: This function is not working as intended when `numpy` is a dependency and already imported.
    """
    if not hasattr(skip, '__iter__') or isinstance(skip, str):
        skip = [skip]
    deps = {}
    with import_package_version(package_name, version, wipe_cache=True):
        package_path = _PACKAGE_PATH / package_name / version
        pkgs = _get_available_packages_in_path(package_path)
        pkgs = [pkg for pkg in pkgs if pkg not in skip]
        # for pkg in pkgs:
        #     _remove_import_from_sys(pkg)
        for pkg in pkgs:
            if pkg in ['numpy', 'numba', 'scipy']:
                # These are being difficult
                deps[pkg] = None
            else:
                # mod = _import_package_from_path(pkg, package_path)
                mod = importlib.import_module(pkg)
                assert mod.__file__.startswith(package_path.as_posix())
                deps[pkg] = mod.__version__ if hasattr(mod, '__version__') else None
    return deps


def _get_available_packages_in_path(path):
    return [pkg.name for pkg in path.glob('*') if (path / pkg).is_dir() \
            and not pkg.name.startswith('_') and not '.' in pkg.name \
            and not pkg.name.endswith('dist-info') and not pkg.name == 'bin']


def _remove_import_from_sys(package_name):
    for mod_name in list(sys.modules):
        if package_name == mod_name or mod_name.startswith(package_name + '.'):
            del sys.modules[mod_name]

def _import_package_from_path(package_name, package_path):
    original_sys_path = sys.path.copy()
    original_sys_meta_path = sys.meta_path.copy()
    for finder_name in sys.meta_path:
        if package_name in str(finder_name):
            sys.meta_path.remove(finder_name)
    sys.path.insert(0, package_path.as_posix())
    module = importlib.import_module(package_name)
    sys.path = original_sys_path
    sys.meta_path = original_sys_meta_path
    return module



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os
import sys
import stat
from subprocess import run, PIPE, CalledProcessError
from pathlib import Path
import warnings

from .fs_methods import make_stat_result, size_expand, _xrdcp_installed

EOS_CELL = 'cern.ch'
default_eos_instance = 'public'

_eos_path = Path('/eos')

_eos_version = -1
_eos_version_int = -1
_eos_arg_for_find = 'find'
_eoscmd_installed = False
if os.name != 'nt':
    try:
        cmd = run(['eos', '--version'], stdout=PIPE, stderr=PIPE)
        # Temporary hack as the eos command wrongly returns 255
        _eoscmd_installed =  cmd.returncode == 0 or cmd.returncode == 255

        if _eoscmd_installed:
            _eos_version = cmd.stdout.decode('UTF-8').strip().split()
            if _eos_version == []:
                _eos_version = cmd.stderr.decode('UTF-8').strip().split()
            _eos_version = _eos_version[1]
            _eos_version_int = _eos_version.split('.')
            _eos_version_int = int(1.e6*int(_eos_version_int[0]) + 1.e3*int(_eos_version_int[1]) \
                                + int(_eos_version_int[2]))
            # The command `eos find` has the wrong behaviour on new machines (running eos 5.2).
            # For this reason, the command `eos oldfind` has been introduced.
            # However, this command does not exist on the old machines, so in that case we
            # default back to `eos find`. This function gives the correct argument for eos find.
            eos_env = {**os.environ, 'EOS_MGM_URL':  f'root://eos{default_eos_instance}.{EOS_CELL}'}
            cmd = run(['eos', 'oldfind'], stdout=PIPE, stderr=PIPE, env=eos_env)
            if cmd.returncode != 255:
                # Command found; we are running on a machine running new eos >= 5.2
                _eos_arg_for_find = 'oldfind'
    except (CalledProcessError, FileNotFoundError):
        _eoscmd_installed = False

_eos_mounted = _eos_path.exists()
eos_accessible = _eos_mounted or _eoscmd_installed or _xrdcp_installed

def _assert_eos_accessible(mess=None):
    if not eos_accessible:
        mess = f" {mess}" if mess is not None else mess
        raise OSError(f"EOS is not installed on your system.{mess}")


def _run_eos(eos_cmds, mgm, _force=False, **kwargs):
    from xaux.fs import _skip_eos_software, _force_eoscmd, _force_xrdcp
    # Try to run the eos command. Returns a bool success and the stdout.
    if _eoscmd_installed and not _skip_eos_software:
        _false_if_stderr_contains = kwargs.pop('_false_if_stderr_contains', None)
        stderr = f"Failed {' '.join(eos_cmds)}.\n"
        eos_env = {**os.environ, 'EOS_MGM_URL': mgm}
        try:
            cmd = run(eos_cmds, stdout=PIPE, stderr=PIPE, env=eos_env)
        except Exception as e:
            stderr += repr(e)
            if _eos_mounted and not _force_eoscmd:
                # We can retry with local FS
                warnings.warn(stderr, RuntimeWarning)
                return False, ''
            else:
                raise OSError(stderr)
        if cmd.returncode == 0:
            stdout = cmd.stdout.decode('UTF-8').strip()
            return True, stdout
        else:
            stderr += cmd.stderr.decode('UTF-8').strip()
            if _false_if_stderr_contains is not None:
                if _false_if_stderr_contains in stderr:
                    return True, False
            if _eos_mounted and not _force_eoscmd:
                # We can retry with local FS
                warnings.warn(stderr, RuntimeWarning)
                return False, ''
            else:
                raise OSError(stderr)
    elif _force:
        raise OSError("The command `eos` is not installed on your system.")
    else:
        return False, None


def is_egroup_member(egroup, verbose=False):
    success, result = _run_eos(['eos', 'member', egroup], _force=True,
                               mgm=f'root://eos{default_eos_instance}.{EOS_CELL}')
    out = result.split()
    is_member = [f.split('=')[1] for f in out if f.startswith('member=')][0] == 'true'
    if verbose:
        mem = 'a' if is_member else 'not a'
        usr = [f.split('=')[1] for f in out if f.startswith('user=')][0]
        egr = [f.split('=')[1] for f in out if f.startswith('egroup=')][0]
        lft = [f.split('=')[1] for f in out if f.startswith('lifetime=')][0]
        print(f"User {usr} is {mem} member of e-group {egr} (lifetime {lft}).")
    return is_member


# Overwrite Path methods
# ======================

def _get_type(path, *args, **kwargs):
    success, result = _run_eos(['eos', 'stat', path.eos_path], mgm=path.mgm,
                        _false_if_stderr_contains='failed to stat', **kwargs)
    if success:
        if not result:
            raise FileNotFoundError
        if result.endswith(' regular file'):
            parts = result.split()
            size = int(parts[parts.index('Size:')+1])
            return stat.S_IFREG, size
        elif result.endswith(' directory'):
            return stat.S_IFDIR, None
        elif result.endswith(' symbolic link'):
            return stat.S_IFLNK, None
        else:
            raise NotImplementedError(f"File type not known. Output:\n{result}")
    return None, None

def _eos_exists(path, *args, **kwargs):
    _assert_eos_accessible("Cannot stat EOS paths.")
    try:
        ftype, _ = _get_type(path, *args, **kwargs)
    except FileNotFoundError:
        return False
    if ftype is not None:
        return True
    return Path(path.eos_path).exists()

def _eos_is_file(path, *args, **kwargs):
    _assert_eos_accessible("Cannot stat EOS paths.")
    try:
        ftype, _ = _get_type(path, *args, **kwargs)
    except FileNotFoundError:
        return False
    if ftype is not None:
        return ftype == stat.S_IFREG
    return Path.is_file(path)

def _eos_is_dir(path, *args, **kwargs):
    _assert_eos_accessible("Cannot stat EOS paths.")
    try:
        ftype, _ = _get_type(path, *args, **kwargs)
    except FileNotFoundError:
        return False
    if ftype is not None:
        return ftype == stat.S_IFDIR
    return Path.is_dir(path)

def _eos_is_symlink(path, *args, **kwargs):
    _assert_eos_accessible("Cannot stat EOS paths.")
    try:
        ftype, _ = _get_type(path, *args, **kwargs)
    except FileNotFoundError:
        return False
    if ftype is not None:
        return ftype == stat.S_IFLNK
    return Path.is_symlink(path)


def _parse_fileinfo(fileinfo, ftype=None, st_size=None):
    data = fileinfo.split()
    stat_dict = {}
    # Get the file type and permissions
    if 'Flags:' in data:
        st_mode = int(data[data.index('Flags:')+1], base=8)
    else:
        st_mode = 0
    if ftype is None:
        ftype = st_mode & 0o770000  # mask to get file type bits
        if ftype == 0:              # parse manually from text
            if data[0] == 'Directory:':
                ftype = stat.S_IFDIR
            elif data[0] == 'File:':
                ftype = stat.S_IFREG
            else:
                raise NotImplementedError(f"Unknown file type {data[0]}.")
    else:
        if st_mode & 0o770000 == 0: # check that it doesn't exist yet in the flags
            st_mode += ftype
        elif ftype != st_mode & 0o770000:
            raise ValueError(f"Provided file type {ftype} and the bit in the flags "
                          + f"{st_mode & 0o770000} do not match!")
    stat_dict['st_mode'] = st_mode
    # Get the size if file or if provided
    if st_size is not None:
        stat_dict['st_size'] = st_size
    elif ftype == stat.S_IFREG and 'Size:' in data:
        stat_dict['st_size'] = size_expand(data[data.index('Size:')+1])
    # Get the other metadata
    if 'Modify:' in data:
        stat_dict['st_mtime'] = float(data[data.index('Timestamp:', data.index('Modify:'))+1])
        stat_dict['st_mtime_ns'] = int(1.e9*stat_dict['st_mtime'])
    if 'Change:' in data:
        stat_dict['st_ctime'] = float(data[data.index('Timestamp:', data.index('Change:'))+1])
        stat_dict['st_ctime_ns'] = int(1.e9*stat_dict['st_ctime'])
    if 'Access:' in data:
        stat_dict['st_atime'] = float(data[data.index('Timestamp:', data.index('Access:'))+1])
        stat_dict['st_atime_ns'] = int(1.e9*stat_dict['st_atime'])
    if 'Birth:' in data:
        stat_dict['st_birthtime'] = float(data[data.index('Timestamp:', data.index('Birth:'))+1])
        stat_dict['st_birthtime_ns'] = int(1.e9*stat_dict['st_birthtime'])
    if 'CUid:' in data:
        stat_dict['st_uid'] = int(data[data.index('CUid:')+1])
    if 'CGid:' in data:
        stat_dict['st_gid'] = int(data[data.index('CGid:')+1])
    if 'Blocksize:' in data:
        stat_dict['st_blksize'] = size_expand(data[data.index('Blocksize:')+1], binary=True)
    # TODO: Not set: st_ino, st_dev, st_nlink, st_flags, st_blocks, st_gen, st_rdev,
    #       st_fstype, st_rsize, st_creator, st_type, st_file_attributes, st_reparse_tag
    return make_stat_result(stat_dict)

if sys.version_info >= (3, 10):
    def _eos_lstat(path, *args, **kwargs):
        _assert_eos_accessible("Cannot stat EOS paths.")
        ftype, st_size = _get_type(path, *args, **kwargs)
        if ftype is None:
            return os.stat(path.as_posix(), follow_symlinks=False)
        elif ftype == stat.S_IFLNK:
            # Special treatment: do NOT follow link
            # Temporary solution: no way to retrieve other info currently
            return make_stat_result({'st_mode': ftype+0o0777, 'st_size': st_size})
        else:
            success, result = _run_eos(['eos', 'fileinfo', path.eos_path], mgm=path.mgm,
                            _false_if_stderr_contains='No such file or directory', **kwargs)
            if not success:
                return os.stat(path.as_posix(), follow_symlinks=False)
            else:
                if not result:
                    raise FileNotFoundError
                return _parse_fileinfo(result, ftype, st_size)

    def _eos_stat(path, *args, follow_symlinks=True, **kwargs):
        _assert_eos_accessible("Cannot stat EOS paths.")
        # The command `eos fileinfo` automatically resolves symlinks
        if not follow_symlinks:
            return _eos_lstat(path, *args, **kwargs)
        success, result = _run_eos(['eos', 'fileinfo', path.eos_path], mgm=path.mgm,
                            _false_if_stderr_contains='No such file or directory', **kwargs)
        if not success:
            return os.stat(path, follow_symlinks=True)
        if not result:
            raise FileNotFoundError
        ftype, _ = _get_type(path, *args, **kwargs)
        return _parse_fileinfo(result, ftype)

else:
    def _eos_lstat(path, *args, **kwargs):
        _assert_eos_accessible("Cannot stat EOS paths.")
        ftype, st_size = _get_type(path, *args, **kwargs)
        if ftype is None:
            return os.lstat(path.as_posix())
        elif ftype == stat.S_IFLNK:
            # Special treatment: do NOT follow link
            # Temporary solution: no way to retrieve other info currently
            return make_stat_result({'st_mode': ftype+0o0777, 'st_size': st_size})
        else:
            success, result = _run_eos(['eos', 'fileinfo', path.eos_path], mgm=path.mgm,
                            _false_if_stderr_contains='No such file or directory', **kwargs)
            if not success:
                return os.lstat(path.as_posix())
            else:
                if not result:
                    raise FileNotFoundError
                return _parse_fileinfo(result, ftype, st_size)

    def _eos_stat(path, *args, **kwargs):
        _assert_eos_accessible("Cannot stat EOS paths.")
        # The command `eos fileinfo` automatically resolves symlinks
        success, result = _run_eos(['eos', 'fileinfo', path.eos_path], mgm=path.mgm,
                            _false_if_stderr_contains='No such file or directory', **kwargs)
        if not success:
            return os.stat(path.as_posix())
        if not result:
            raise FileNotFoundError
        ftype, _ = _get_type(path, *args, **kwargs)
        return _parse_fileinfo(result, ftype)


def _eos_touch(path, mode=0o666, exist_ok=True, **kwargs):
    _assert_eos_accessible("Cannot touch EOS paths.")
    success, result = _run_eos(['eos', 'touch', path.eos_path], mgm=path.mgm, **kwargs)
    if success:
        return result
    return Path.touch(path, mode=mode, exist_ok=exist_ok)

def _eos_unlink(path, missing_ok=False, **kwargs):
    _assert_eos_accessible("Cannot unlink EOS paths.")
    if not path.is_symlink(**kwargs) and path.is_dir(**kwargs):
        raise IsADirectoryError(f"{path} is a directory.")
    success, result = _run_eos(['eos', 'rm', path.eos_path], mgm=path.mgm, **kwargs)
    if success:
        return result
    return Path.unlink(path, missing_ok=missing_ok)

def _eos_mkdir(path, mode=0o777, parents=False, exist_ok=False, **kwargs):
    _assert_eos_accessible("Cannot rmdir EOS paths.")
    success, result = _run_eos(['eos', 'mkdir', path.eos_path], mgm=path.mgm, **kwargs)
    if success:
        return result
    return Path.mkdir(path, mode=mode, parents=parents, exist_ok=exist_ok)

def _eos_rmdir(path, *args, **kwargs):
    _assert_eos_accessible("Cannot rmdir EOS paths.")
    if path.is_symlink(*args, **kwargs) or not path.is_dir(*args, **kwargs):
        raise NotADirectoryError(f"{path} is not a directory.")
    success, result = _run_eos(['eos', 'rmdir', path.eos_path], mgm=path.mgm, **kwargs)
    if success:
        return result
    return Path.rmdir(path)


# Overwrite FsPath methods
# ======================

def _eos_symlink_to(path, def_cls, target, target_is_directory=False, **kwargs):
    _assert_eos_accessible("Cannot create symlinks on EOS paths.")
    success, result = _run_eos(['eos', 'ln', '-fns', path.eos_path, target.as_posix()],
                               mgm=path.mgm, **kwargs)
    if success:
        return result
    from xaux.fs import FsPath
    return FsPath.symlink_to(path, target, target_is_directory)

def _eos_rmtree(path, def_cls, *args, **kwargs):
    _assert_eos_accessible("Cannot rmtree EOS paths.")
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory.")
    success, result = _run_eos(['eos', 'rm', '-r', path.eos_path], mgm=path.mgm, **kwargs)
    if success:
        return result
    return def_cls.rmtree(path, *args, **kwargs)

def _eos_size(path, def_cls, *args, **kwargs):
    _assert_eos_accessible("Cannot get size of EOS paths.")
    if not path.is_file():
        return 0
    success, result = _run_eos(['eos', 'stat', path.eos_path], mgm=path.mgm, **kwargs)
    if success:
        return int(result.split()[3])
    return def_cls.size(path, *args, **kwargs)




# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import os
from shutil import copy2, copytree
from subprocess import run, PIPE
import warnings

from .fs import FsPath
from .afs import AfsPath, _afs_mounted
from .eos import EosPath
from .eos_methods import _eos_mounted, _xrdcp_installed, _eoscmd_installed, _eos_version, _eos_version_int


# TODO:
#   - xrdcp: increase efficiency by using a list of files
#   - xrdcp: check if recursive still needs to be done manually
#   - xrdcp: can we parallelize?
#   - xrdcp and eos: implement follow_symlinks
#   - test if symlinks behave correctly
#   - EOS is very very slow


# We want to mimic the bash commands as much as possible
# cp file1 file2            file2 does not exist:         copies file1 into new file2
# cp file1 file2            file2 does exist:             copies file1 into file2 (overwrites file2)
# cp file1 file2 file3                                    error  "cp: target 'file3' is not a directory"
# cp file1 link1            link points to other file:    copies file1 into link1 (overwrites the contents of the linked file without renaming it)
# cp file1 link1            link points to file1:         error "cp: 'file1' and 'link1' are the same file"
# cp file1 file2 link1                                    error  "cp: target 'file3' is not a directory"
# cp file1 dir1                                           copies file1 into dir1 (potentially overwrites)
# cp file1 file2 ... dir1                                 copies files into dir1 (potentially overwrites)
# cp file1 file2 ... link1  link points to dir1:          copies files into dir1 (potentially overwrites)
# cp dir1 file1                                           error  "cp: -r not specified; omitting directory 'dir1'"
# cp -r dir1 file1                                        error  "cp: cannot overwrite non-directory 'file1' with directory 'dir1'"
# cp -r dir1 dir2           dir2 does not exist:          copies dir1 as new dir2
# cp -r dir1 dir2           dir2 does exist:              copies dir1 inside dir2
# links are treated as the file/dir they point to, unless -P is specified, in which case (only for links as source argument) they are treated as separate files


# If follow_symlinks=False, same as cp -P (hence a link is copied instead of its contents)
def cp(*args, recursive=False, follow_symlinks=True, **kwargs):
    # print("COPY"); import time; t_start = time.time(); t_prev = t_start
    if len(args) < 2:
        return
    this_stdout = ""
    this_stderr = ""
    args = [FsPath(arg).expanduser() for arg in args]
    target = args[-1].resolve()
    sources = args[:-1]
    # t_new = time.time(); print(f"{target=}, {sources=}  ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new

    # Renaming is only allowed if there is a single source
    if len(sources) > 1:
        if not target.is_dir():
            raise OSError(f"cp: target '{target}' is not a directory")

    # Omitting directories if -r is not specified
    if not recursive:
        new_sources = []
        for src in sources:
            if src.is_dir():
                this_stdout += f"cp: -r not specified; omitting directory '{src}'"
            else:
                new_sources.append(src)
        sources = new_sources

    # Match target for each source and verify the syntax
    # t_new = time.time(); print(f"Before loop ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new
    sources_targets = _loop_sources_and_verify(sources, target)
    # t_new = time.time(); print(f"After loop ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new
    # print(f"{sources_targets=}")

    # Files on EOS or AFS need a special treatment
    if isinstance(target, EosPath):
        eos_sources_targets = sources_targets
        afs_sources_targets = []
        sources_targets = []
    elif isinstance(target, AfsPath):
        eos_sources_targets = [f for f in sources_targets if isinstance(f[0], EosPath)]
        afs_sources_targets = [f for f in sources_targets if not isinstance(f[0], EosPath)]
        sources_targets = []
    else:
        eos_sources_targets = [f for f in sources_targets if isinstance(f[0], EosPath)]
        afs_sources_targets = [f for f in sources_targets if isinstance(f[0], AfsPath)]
        sources_targets     = [f for f in sources_targets if not isinstance(f[0], EosPath)
                                                         and not isinstance(f[0], AfsPath)]

    # t_new = time.time(); print(f"Setup done ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new
    # Do the copy
    stdout, stderr = _cp_regular(sources_targets, follow_symlinks, **kwargs)
    this_stdout += stdout
    this_stderr += stderr
    # t_new = time.time(); print(f"Regular done ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new
    stdout, stderr = _cp_afs(afs_sources_targets, follow_symlinks, **kwargs)
    this_stdout += stdout
    this_stderr += stderr
    # t_new = time.time(); print(f"AFS done ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new
    stdout, stderr = _cp_eos(eos_sources_targets, follow_symlinks, **kwargs)
    this_stdout += stdout
    this_stderr += stderr
    # t_new = time.time(); print(f"EOS done ({int(1e3*(t_new-t_prev))}ms)"); t_prev = t_new

    # Raise an exception if some files could not be copied
    if this_stderr:
       raise OSError(this_stderr)

    return this_stdout


# If follow_symlinks=False, same as cp -P (hence a link is copied instead of its contents)
def mv(*args, follow_symlinks=True, **kwargs):
    stdout = cp(*args, recursive=True, follow_symlinks=follow_symlinks, **kwargs)
    # If we got here, then the copy was successful
    for arg in args[:-1]:
        if arg.is_dir():
            arg.rmtree()
        else:
            arg.unlink()
    return stdout


# ===================
# Low-level functions
# ===================

def _loop_sources_and_verify(sources, target):
    src_target_recursive = []
    for src in sources:
        if src.is_dir():
            if target.exists():
                if not target.is_dir():
                    raise OSError(f"cp: cannot overwrite non-directory '{target}' with directory '{src}'\n")
                this_target = target / src.name
            elif target.parent.exists():
                # This includes a potential renaming
                this_target = target
            else:
                raise OSError(f"cp: cannot create directory '{target}': No such file or directory\n")
            if src in this_target.parents:
                raise OSError(f"cp: cannot copy a directory, '{src}', into itself, '{target}'\n")
            src_target_recursive.append([src, this_target, True])

        else:
            if target.exists():
                if target.is_dir():
                    this_target = target / src.name
                else:
                    # We are overwriting the target
                    this_target = target
            elif target.parent.exists():
                # This includes a potential renaming
                this_target = target
            else:
                raise OSError(f"cp: cannot create regular file '{target}': No such file or directory")
            if this_target == src:
                raise OSError(f"cp: '{src}' and '{this_target}' are the same file\n")
            src_target_recursive.append([src, this_target, False])

    return src_target_recursive


def _cp_regular(sources_targets, follow_symlinks, **kwargs):
    stdout = ""
    stderr = ""
    for src, target, recursive in sources_targets:
        if recursive:
            cmd_mess  = f"copytree({src}, {target}, symlinks={not follow_symlinks}, "
            cmd_mess += "copy_function=copy2)"
            try:
                this_stdout = copytree(src.as_posix(), target.as_posix(), symlinks=not follow_symlinks, copy_function=copy2)
                if this_stdout is not None:
                    stdout += this_stdout
                # Verify the files exist
                if not target.exists():
                    stderr += f"Failed {cmd_mess}:\n"
                    stderr += f"   Target {target} does not exist.\n"
            except Exception as e:
                stderr += f"Failed {cmd_mess}:\n"
                stderr += f"   {str(e)}\n"

        else:
            cmd_mess = f"copy2({src}, {target}, follow_symlinks={follow_symlinks})"
            try:
                this_stdout = copy2(src.as_posix(), target.as_posix(), follow_symlinks=follow_symlinks)
                if this_stdout is not None:
                    stdout += this_stdout
                # Verify the files exist
                if not target.exists():
                    stderr += f"Failed {cmd_mess}:\n"
                    stderr += f"   Target {target} does not exist.\n"
            except Exception as e:
                stderr += f"Failed {cmd_mess}:\n"
                stderr += f"   {str(e)}\n"

    return stdout, stderr  # An empty error message means success


def _cp_afs(sources_targets, follow_symlinks, method=None, **kwargs):
    from xaux.fs import _skip_afs_software, _force_xrdcp
    this_stdout = ""
    this_stderr = ""
    if _skip_afs_software:
        assert not _force_xrdcp

    if sources_targets and method:
        if method not in ['mount', 'xrdcp']:
            raise ValueError(f"Invalid method '{method}' for AFS copy. "
                           + f"Use 'mount' or 'xrdcp'.")
        elif method == 'xrdcp':
            if not _xrdcp_installed:
                raise RuntimeError("xrdcp is not installed. Cannot copy "
                                + "files to/from AFS using xrdcp.")
            elif _skip_afs_software:
                raise RuntimeError("Cannot copy files to/from AFS using xrdcp "
                                + "because AFS software is skipped.")
        elif method == 'mount':
            if not _afs_mounted:
                raise RuntimeError("AFS is not mounted. Cannot copy files "
                                + "to/from AFS using mount.")
            elif _force_xrdcp:
                raise RuntimeError("Cannot copy files to/from AFS using mount "
                                + "because xrdcp is forced.")

    # We first try to use xrdcp
    if sources_targets and _xrdcp_installed and not _skip_afs_software \
    and method != 'mount':
        sources_targets, stdout, stderr = _cp_xrdcp(sources_targets)
        this_stdout += stdout
        this_stderr += stderr

    # If xrdcp failed, we try to use the AFS mout
    regular_stderr = ""
    if sources_targets and method != 'xrdcp':
        if _force_xrdcp:
            this_stderr += "Skipping AFS mount.\n"
            this_stderr += "Failed to copy files to AFS.\n"
            return this_stdout, this_stderr
        if not _afs_mounted:
            this_stderr += "No AFS mount available.\n"
            this_stderr += "Failed to copy files to AFS.\n"
            return this_stdout, this_stderr
        regular_stdout, regular_stderr = _cp_regular(sources_targets, follow_symlinks)
        this_stdout += regular_stdout

    if not regular_stderr:
        # All files are copied successfully
        if this_stderr:
            # Just raise a warning that some earlier steps failed
            warnings.warn(this_stderr, RuntimeWarning)
        return this_stdout, ""  # An empty error message means success
    else:
        return this_stdout, this_stderr + regular_stderr


def _cp_eos(sources_targets, follow_symlinks, method=None, **kwargs):
    from xaux.fs import _skip_eos_software, _force_eoscmd, _force_xrdcp
    this_stdout = ""
    this_stderr = ""
    if _skip_eos_software:
        assert not _force_eoscmd
        assert not _force_xrdcp

    if sources_targets and method:
        if method not in ['mount', 'xrdcp', 'eoscmd']:
            raise ValueError(f"Invalid method '{method}' for EOS copy. "
                           + f"Use 'mount', 'xrdcp' or 'eoscmd'.")
        elif method == 'xrdcp':
            if not _xrdcp_installed:
                raise RuntimeError("xrdcp is not installed. Cannot copy "
                                 + "files to/from EOS using xrdcp.")
            elif _skip_eos_software:
                raise RuntimeError("Cannot copy files to/from EOS using xrdcp "
                                 + "because EOS software is skipped.")
            elif _force_eoscmd:
                raise RuntimeError("Cannot copy files to/from EOS using xrdcp "
                                 + "because eoscmd is forced.")
        elif method == 'eoscmd':
            if not _eoscmd_installed:
                raise RuntimeError("eoscmd is not installed. Cannot copy "
                                 + "files to/from EOS using eoscmd.")
            elif _skip_eos_software:
                raise RuntimeError("Cannot copy files to/from EOS using eoscmd "
                                 + "because EOS software is skipped.")
            elif _force_xrdcp:
                raise RuntimeError("Cannot copy files to/from EOS using eoscmd "
                                 + "because xrdcp is forced.")
        elif method == 'mount':
            if not _eos_mounted:
                raise RuntimeError("EOS is not mounted. Cannot copy files "
                                 + "to/from EOS using mount.")
            elif _force_xrdcp or _force_eoscmd:
                raise RuntimeError("Cannot copy files to/from EOS using mount "
                                 + "because xrdcp or eoscmd are forced.")

    # We first try to use xrdcp
    if sources_targets and _xrdcp_installed and not _force_eoscmd and not _skip_eos_software \
    and method != 'mount' and method != 'eoscmd':
        sources_targets, stdout, stderr = _cp_xrdcp(sources_targets)
        this_stdout += stdout
        this_stderr += stderr

    # If xrdcp failed, we try to use eos
    if sources_targets and _eoscmd_installed and not _force_xrdcp and not _skip_eos_software \
    and method != 'mount' and method != 'xrdcp':
        sources_targets, stdout, stderr = _cp_eoscmd(sources_targets)
        this_stdout += stdout
        this_stderr += stderr

    # If eos also failed, we try using the EOS mount
    regular_stderr = ""
    if sources_targets and method != 'xrdcp' and method != 'eoscmd':
        if _force_xrdcp or _force_eoscmd:
            this_stderr += "Skipping EOS mount.\n"
            this_stderr += "Failed to copy files to EOS.\n"
            return this_stdout, this_stderr
        if not _eos_mounted:
            this_stderr += "No EOS mount available.\n"
            this_stderr += "Failed to copy files to EOS.\n"
            return this_stdout, this_stderr
        regular_stdout, regular_stderr = _cp_regular(sources_targets, follow_symlinks)
        this_stdout += regular_stdout

    if not regular_stderr:
        # All files are copied successfully
        if this_stderr:
            # Just raise a warning that some earlier steps failed
            warnings.warn(this_stderr, RuntimeWarning)
        return this_stdout, ""  # An empty error message means success
    else:
        return this_stdout, this_stderr + regular_stderr


def _cp_xrdcp(sources_targets):
    from xaux.fs import _xrdcp_use_ipv4
    stderr = ""
    stdout = ""
    # Option --xattr is not supported on all machines
    # f'--sources {min(len(path_src), 15)}' f'--streams {min(len(path_src), 15)}' f'--parallel {min(len(path_src), 15)}'
    opts = ['--cksum', 'adler32', '--force', '--nopbar', '--rm-bad-cksum']
    if _xrdcp_use_ipv4:
        env = {'env': {**os.environ, 'XRD_NETWORKSTACK': 'IPv4'}}
    else:
        env = {}

    # TODO: increase efficiency by using a list of files
    # path_src = [src.eos_path_full if isinstance(src, EosPath) else src.as_posix() for src in sources]
    # infiles = path_src
    # if len(path_src) > 1:
    #     # Multiple files are copied using a list of files
    #     _temp = FsPath(f"list_of_files_{ranID()}").resolve()
    #     with _temp.open('w') as fid:
    #         for arg in path_src[:-1]:
    #             fid.write(arg + '\n')
    #     infiles = ['--infiles', _temp.as_posix()]

    cmd_data = []
    for src, target, recursive in sources_targets:
        if recursive:
            if not isinstance(target, EosPath) and not isinstance(target, AfsPath):
                # Not an XRootD path so -r works
                path_src = src.eos_path_full if isinstance(src, EosPath) else src.as_posix()
                path_target = target.eos_path_full if isinstance(target, EosPath) else target.as_posix()
                cmd_data.append(['xrdcp', '-r', *opts, path_src, path_target, src, target, True])
            else:
                # Manually walk through the directory
                for new_src in src.rglob('*'):
                    if new_src.is_dir():
                        # xrdcp automatically creates parent directories
                        continue
                    path_src = new_src.eos_path_full if isinstance(new_src, EosPath) else new_src.as_posix()
                    new_target = target / new_src.relative_to(src)
                    path_target = new_target.eos_path_full if isinstance(new_target, EosPath) else new_target.as_posix()
                    cmd_data.append(['xrdcp', *opts, path_src, path_target, new_src, new_target, False])

        else:
            path_src = src.eos_path_full if isinstance(src, EosPath) else src.as_posix()
            path_target = target.eos_path_full if isinstance(target, EosPath) else target.as_posix()
            cmd_data.append(['xrdcp', *opts, path_src, path_target, src, target, False])

    for this_cmd_data in cmd_data:
        src = this_cmd_data[-3]
        target = this_cmd_data[-2]
        this_cmd = this_cmd_data[:-3]
        cmd_mess = ' '.join(this_cmd)
        try:
            this_stderr = ""
            cmd = run(this_cmd, stdout=PIPE, stderr=PIPE, **env)
            if cmd.returncode == 0:
                # Verify the files exist
                if not target.exists():
                    this_stderr += f"Failed {cmd_mess}:\n"
                    this_stderr += f"   Target {target} does not exist.\n"
            else:
                this_stderr += f"Failed {cmd_mess}:\n"
                this_stderr += f"   {cmd.stderr.decode('UTF-8').strip()}\n"
            this_stdout = cmd.stdout.decode('UTF-8').strip()
            stderr += this_stderr
            if this_stdout:
                stdout += f"{this_stdout}\n"
                if this_stderr:
                    stderr += f"   stdout: {this_stdout}\n"
        except Exception as e:
            stderr += f"Failed {cmd_mess}:\n"
            stderr += f"   {str(e)}\n"
        # else:
        #     # Clean up the temporary file
        #     if infiles[0] == '--infiles':
        #         _temp.unlink()

    sources_targets = [f[-3:] for f in cmd_data if not f[-2].exists()]

    return sources_targets, stdout, stderr


def _cp_eoscmd(sources_targets):
    stderr = ""
    stdout = ""
    cmd_data = []
    for src, target, recursive in sources_targets:
        if recursive:
            if _eos_version_int >= 9000000: # 5002021 DOES NOT WORK
                # -r works
                path_src = src.eos_path if isinstance(src, EosPath) else src.as_posix()
                if src.is_dir() and path_src[-1] != '/':
                    path_src = path_src + '/'
                path_target = target.eos_path if isinstance(target, EosPath) else target.as_posix()
                if target.is_dir() and path_target[-1] != '/':
                    path_target = path_target + '/'
                cmd_data.append(['eos', 'cp', '-r', path_src, path_target, src, target, True])
            else:
                # Manually walk through the directory
                for new_src in src.rglob('*'):
                    if new_src.is_dir():
                        new_src.mkdir(parents=True, exist_ok=True)
                        continue
                    path_src = new_src.eos_path if isinstance(new_src, EosPath) else new_src.as_posix()
                    if new_src.is_dir() and path_src[-1] != '/':
                        path_src = path_src + '/'
                    new_target = target / new_src.relative_to(src)
                    path_target = new_target.eos_path if isinstance(new_target, EosPath) else new_target.as_posix()
                    if new_target.is_dir() and path_target[-1] != '/':
                        path_target = path_target + '/'
                    cmd_data.append(['eos', 'cp', path_src, path_target, new_src, new_target, False])

        else:
            path_src = src.eos_path if isinstance(src, EosPath) else src.as_posix()
            if src.is_dir() and path_src[-1] != '/':
                path_src = path_src + '/'
            path_target = target.eos_path if isinstance(target, EosPath) else target.as_posix()
            if target.is_dir() and path_target[-1] != '/':
                path_target = path_target + '/'
            cmd_data.append(['eos', 'cp', path_src, path_target, src, target, False])

    for this_cmd_data in cmd_data:
        src = this_cmd_data[-3]
        target = this_cmd_data[-2]
        this_cmd = this_cmd_data[:-3]
        cmd_mess = ' '.join(this_cmd)
        if isinstance(target, EosPath):
            mgm = target.mgm
        else:
            mgm = src.mgm
        eos_env = {**os.environ, 'EOS_MGM_URL': mgm}
        try:
            this_stderr = ""
            cmd = run(this_cmd, stdout=PIPE, stderr=PIPE, env=eos_env)
            if cmd.returncode == 0:
                # Verify the files exist
                if not target.exists():
                    this_stderr += f"Failed {cmd_mess}:\n"
                    this_stderr += f"   Target {target} does not exist.\n"
            else:
                this_stderr += f"Failed {cmd_mess}:\n"
                this_stderr += f"   {cmd.stderr.decode('UTF-8').strip()}\n"
            this_stdout = cmd.stdout.decode('UTF-8').strip()
            stderr += this_stderr
            if this_stdout:
                stdout += f"{this_stdout}\n"
                if this_stderr:
                    stderr += f"   stdout: {this_stdout}\n"
        except Exception as e:
            stderr += f"Failed {cmd_mess}:\n"
            stderr += f"   {str(e)}\n"

    sources_targets = [f[-3:] for f in cmd_data if not f[-2].exists()]

    return sources_targets, stdout, stderr



# copyright ############################### #
# This file is part of the Xaux package.    #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import functools


class ClassProperty:
    """Descriptor to define class properties.
    Similar to the built-in property, but for classes instead of instances.
    - Contrary to a regular property, a __set__ or __delete__ call on the
      owner class would not be intercepted. For this reason, it is necessary
      to use a dedicated metaclass, the ClassPropertyMeta, to intercept these
      calls, even when no setter or deleter is defined (as otherwise the
      attribute would not be read-only and could still be overwritten).
    - The ClassProperty class keeps a registry of ClassProperties for each
      class, which is accessible with the 'get_properties' method.
    - Whenever a class has ClassProperties, a ClassPropertyAccessor named
      'classproperty' will be attached to it, providing an attribute-like
      interface to the ClassProperty attributes of a class for introspection.
      Use like ?MyClass.classproperty.my_class_property to get the introspect.
    - An important caveat is that regular class attributes do not always behave
      as expected when inherited, which might be an issue when a ClassProperty
      uses such a regular class attribute (for instance as the private attribute
      it is encapsulating). Indeed, when the parent has a class attribute '_prop'
      it will not be copied unto the child, and any ClassProperty.setter applied
      on the child will inevitably update the parent's attributes as well. To
      handle this, one can define a dict '_classproperty_dependencies' in the
      class to declare all dependent regular class attributes and their initial
      values. The 'ClassPropertyMeta' then copies these attributes to the child.

    Example usage:

    class MyClass(metaclass=ClassPropertyMeta):
        _classproperty_dependencies = {
            '_my_classproperty': 0
        }

        @ClassProperty
        def my_class_property(cls):
            return cls._my_classproperty

        @my_class_property.setter
        def my_class_property(cls, value):
            cls._my_classproperty = value

        @my_class_property.deleter
        def my_class_property(cls):
            cls._my_classproperty = 0
    """

    _registry = {}  # Registry to store ClassProperty names for each class

    @classmethod
    def get_properties(cls, owner, parents=True):
        """Return the ClassProperty attributes of a class, optionally including those of
        its parents."""
        if not parents:
            return cls._registry.get(owner, {})
        else:
            return {name: prop for parent in owner.__mro__
                         for name, prop in cls._registry.get(parent, {}).items()}

    def __repr__(self):
        """Return repr(self)."""
        return f"<ClassProperty at {hex(id(self))}>"

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        """Initialize self. See help(type(self)) for accurate signature."""
        self._fget = fget
        self._fset = fset
        self._fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __set_name__(self, owner, name):
        """Method to set name of a ClassProperty. Also asserts that the correct metaclass
        is used, adds the property to the registry, and creates default getter, setter, and
        deleter functions."""
        self.name = name
        self.owner = owner
        # Check if we have the correct metaclass
        self._assert_metaclass()
        # Add the property name to the registry for the class
        if owner not in ClassProperty._registry:
            ClassProperty._registry[owner] = {}
        ClassProperty._registry[owner][name] = self
        # Create default getter, setter, and deleter
        if self.fget is None:
            def _getter(this_owner):
                raise AttributeError(f"Unreadable attribute '{name}' of {this_owner.__name__} "
                                    + "class!")
            self._fget = _getter
        if self.fset is None:
            def _setter(this_owner, value):
                raise AttributeError(f"ClassProperty '{name}' of '{this_owner.__name__}' class "
                                    + "has no setter")
            self._fset = _setter
        if self.fdel is None:
            def _deleter(this_owner):
                raise AttributeError(f"ClassProperty '{name}' of '{this_owner.__name__}' class "
                                    + "has no deleter")
            self._fdel = _deleter
        # Attach an accessor to the parent class to inspect ClassProperties
        if not 'classproperty' in owner.__dict__:
            owner.classproperty = ClassPropertyAccessor()
        elif not isinstance(owner.__dict__['classproperty'], ClassPropertyAccessor):
            raise TypeError(f"Class '{owner.__name__}' already has an attribute 'classproperty' "
                          + f"of type {type(owner.__dict__['classproperty']).__name__}! This is "
                          + "incompatible with the ClassProperty descriptor.")

    @property
    def fget(self):
        """Return the getter function of the ClassProperty."""
        # The fget function can only be set at initialisation, or with the getter method.
        return self._fget

    @property
    def fset(self):
        """Return the setter function of the ClassProperty."""
        # The fset function can only be set at initialisation, or with the setter method.
        return self._fset

    @property
    def fdel(self):
        """Return the deleter function of the ClassProperty."""
        # The fdel function can only be set at initialisation, or with the deleter method.
        return self._fdel

    def __get__(self, instance, owner):
        """Return an attribute of owner."""
        if owner is None:
            owner = type(instance)
        try:
            return self.fget(owner)
        except ValueError:
            # Return a fallback if initialisation fails
            return None

    def __set__(self, owner, value):
        """Set a class attribute of owner to value."""
        self.fset(owner, value)

    def __delete__(self, owner):
        """Delete an attribute of owner."""
        self.fdel(owner)

    def getter(self, fget):
        """Decorator to set the ClassProperty's getter fget."""
        self._fget = fget
        return self

    def setter(self, fset):
        """Decorator to set the ClassProperty's setter fset. Need ClassPropertyMeta for this."""
        # We check the metaclass, even though this is already done in __set_name__, for the
        # rare edge case where one would call the setter method directly after the class
        # definition.
        self._assert_metaclass()
        self._fset = fset
        return self

    def deleter(self, fdel):
        """Decorator to set the ClassProperty's deleter fdel. Need ClassPropertyMeta for this."""
        self._assert_metaclass()
        self._fdel = fdel
        return self

    def _assert_metaclass(self):
        # Verify that the metaclass is a subclass of ClassPropertyMeta, needed for fset and fdel
        if hasattr(self, 'owner'):
            if ClassPropertyMeta not in type(self.owner).__mro__:
                raise TypeError(f"Class '{self.owner.__name__}' must have ClassPropertyMeta "
                            + f"as a metaclass to be able to use ClassProperties!")


class ClassPropertyMeta(type):
    """Metaclass for classes with ClassProperty attributes.
    This metaclass intercepts __setattr__ and __delattr__ calls, such that
    setting or deleting a ClassProperty attribute on the class itself works
    as expected. Both functions are defined at the metaclass level (to be
    able to intercept calls to the class properties), but are also defined/
    overwritten at the class level to be able to intercept the same calls
    but on instances.
    """

    def __new__(cls, name, bases, data, new_class=None):
        """Define the new class. By allowing the new_class argument, we can
        easily combine the metaclass with other metaclasses that define the
        __new__ method.
        """

        if new_class is None:
            new_class = type.__new__(cls, name, bases, data)

        # Overwrite the __setattr__ method in the class
        original_setattr = new_class.__dict__.get('__setattr__', None)
        def __setattr__(self, key, value):
            """The __setattr__ method is expanded by the ClassProperty metaclass."""
            # This is the __setattr__ method when called on an INSTANCE of the class.

            this_cls = type(self)
            # Check if the attribute is a ClassProperty
            for parent in this_cls.__mro__:
                if key in parent.__dict__ and isinstance(parent.__dict__[key], ClassProperty):
                    return parent.__dict__[key].__set__(this_cls, value)
            # If not, call the original __setattr__ method.
            # However, there is still a potential issue. If we are setting a regular class
            # attribute (like the underscore variable that goes with the ClassProperty),
            # we need to set it on the class (this_cls) instead of on this instance (self);
            # otherwise the attribute would be created newly on the instance (desynced from
            # the class attribute). So we check if this attribute exists in the class dict.
            # If yes, call the metaclass __setattr__ method to set it on the class. If not,
            # call the original __setattr__ method to set it on the instance.
            # Final caveat: if the attribute is a descriptor (like a property), we DO want
            # to set it on the instance (as this is how descriptors are implemented). We
            # can recognise this by checking if the attribute has a __get__ method.
            if key in this_cls.__dict__ and not hasattr(this_cls.__dict__[key], '__get__'):
                # Set the attribute on the class
                return super(ClassPropertyMeta, this_cls).__setattr__(key, value)
            else:
                # Set the attribute on the instance
                if original_setattr is not None:
                    return original_setattr(self, key, value)
                else:
                    # We have to call super on new_class, i.e. the class this method is
                    # attached to. Otherwise we will end up in infinite loops in case of
                    # inheritance.
                    return super(new_class, self).__setattr__(key, value)
        new_class.__setattr__ = functools.wraps(ClassPropertyMeta.__setattr__)(__setattr__)

        # Overwrite the __delattr__ method in the class
        original_delattr = new_class.__dict__.get('__delattr__', None)
        def __delattr__(self, key):
            """The __delattr__ method is expanded by the ClassProperty metaclass."""
            # This is the __delattr__ method when called on an INSTANCE of the class.

            this_cls = type(self)
            # Check if the attribute is a ClassProperty
            for parent in this_cls.__mro__:
                if key in parent.__dict__ and isinstance(parent.__dict__[key], ClassProperty):
                    return parent.__dict__[key].__delete__(this_cls)
            # If not, call the original __delattr__ method.
            # The logic here is the same as in the __setattr__ method above.
            if key in this_cls.__dict__ and not hasattr(this_cls.__dict__[key], '__get__'):
                return super(ClassPropertyMeta, this_cls).__delattr__(key)
            else:
                if original_delattr is not None:
                    return original_delattr(self, key)
                else:
                    return super(new_class, self).__delattr__(key)
        new_class.__delattr__ = functools.wraps(ClassPropertyMeta.__delattr__)(__delattr__)

        # Get all dependencies that are used by the ClassProperties from the parents into the class
        for parent in new_class.__mro__:
            if '_classproperty_dependencies' in parent.__dict__:
                for key, value in parent.__dict__['_classproperty_dependencies'].items():
                    setattr(new_class, key, value)

        return new_class

    # Define __setattr__  at the metaclass level to intercept setter calls on the class
    def __setattr__(cls, key, value):
        """Set an attribute on the instance or the class, handling ClassProperty if needed."""
        # This is the __setattr__ method when called on the class ITSELF (not an instance).

        # Check if the attribute is a ClassProperty
        for parent in cls.__mro__:
            if key in parent.__dict__ and isinstance(parent.__dict__[key], ClassProperty):
                return parent.__dict__[key].__set__(cls, value)
        # If not, call the original __setattr__ method
        return super(ClassPropertyMeta, cls).__setattr__(key, value)

    # Define __delattr__  at the metaclass level to intercept deleter calls on the class
    def __delattr__(cls, key):
        """Delete an attribute from the instance or the class, handling ClassProperty if needed."""
        # This is the __delattr__ method when called on the class ITSELF (not an instance).

        # Check if the attribute is a ClassProperty
        for parent in cls.__mro__:
            if key in parent.__dict__ and isinstance(parent.__dict__[key], ClassProperty):
                return parent.__dict__[key].__delete__(cls)
        # If not, call the original __delattr__ method
        return super(ClassPropertyMeta, cls).__delattr__(key)


class ClassPropertyAccessor:
    """Helper class to introspect the ClassProperty attributes of a class and
    their docstrings."""

    def __repr__(self):
        """Return repr(self)."""
        return f"<ClassPropertyAccessor at {hex(id(self))}>"

    def __get__(self, instance, owner):
        """Return a ClassPropertyDict object for the owner class."""
        return ClassPropertyDict(owner)


class ClassPropertyDict:
    """Helper class to provide an attribute-like interface to the ClassProperty
    attributes of a class. This way, one can do e.g. ?MyClass.classproperty.cprop1
    to get the introspect info of cprop1."""

    def __init__(self, owner):
        """Initialize self. See help(type(self)) for accurate signature."""
        self.owner = owner
        self._cprops = ClassProperty.get_properties(owner, parents=True)

    @property
    def names(self):
        """Return the names of the ClassProperty attributes of the owner class."""
        return tuple(self._cprops.keys())

    def __repr__(self):
        """Return repr(self)."""
        num_props = len(self)
        props = "ClassProperties" if num_props != 1 else "ClassProperty"
        return f"<ClassPropertyDict on class {self.owner.__name__} ({num_props} "\
             + f"{props}) at {hex(id(self))}>"

    def __iter__(self):
        """Implement iter(self)."""
        return self._cprops.__iter__()

    def keys(self):
        """A set-like object providing a view on the ClassProperty names."""
        return self._cprops.keys()

    def values(self):
        """A set-like object providing a view on the ClassProperties."""
        return self._cprops.values()

    def items(self):
        """A set-like object providing a view on the ClassProperties and their names."""
        return self._cprops.items()

    def __len__(self):
        """Return len(self)."""
        return len(self._cprops)

    def __contains__(self, key):
        """Return key in self."""
        return key in self._cprops

    def __getattr__(self, name):
        """Access the ClassProperty attributes of the owner class."""
        if name not in self._cprops:
            raise AttributeError(f"Class '{self.owner.__name__}' has no ClassProperty '{name}'")
        return self._cprops.get(name)



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .general import _pkg_root, __version__
from .fs import FsPath, LocalPath, EosPath, AfsPath, afs_accessible, eos_accessible, is_egroup_member, cp, mv
from .dev_tools import import_package_version  # Stub to get dev_tools in the namespace
from .tools import singleton, ClassProperty, ClassPropertyMeta, timestamp, ranID, system_lock, get_hash, \
                   ProtectFile, ProtectFileError

_tempdir_cache = None
def __getattr__(name):
    global _tempdir_cache
    if name == "tempdir":
        if not _tempdir_cache:
            from .fs.temp import _tempdir
            _tempdir_cache = _tempdir
        return FsPath(_tempdir_cache)
    raise AttributeError(f"module {__name__} has no attribute {name}")



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .release_tools import make_release, make_release_branch, rename_release_branch
from .package_manager import import_package_version, install_package_version, get_package_versions, \
                             get_latest_package_version, get_package_dependencies, \
                             get_package_version_dependencies
from .gh import assert_git_repo, assert_git_repo_name, assert_gh_installed, assert_poetry_installed, \
                git_assert_working_tree_clean, git_current_branch, git_rename_current_branch, \
                git_switch, git_add, git_commit, git_pull, git_push, git_make_tag, gh_pr_create, \
                gh_pr_list, gh_pr_merge, gh_release_create, poetry_bump_version, poetry_get_version, \
                poetry_get_expected_version, poetry_publish, GitError, GhError, PoetryError



# copyright ############################### #
# This file is part of the Xaux Package.    #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .general_tools import timestamp, ranID, system_lock, get_hash
from .function_tools import count_arguments, count_required_arguments, count_optional_arguments, \
                            has_variable_length_arguments, has_variable_length_positional_arguments, \
                            has_variable_length_keyword_arguments
from .singleton import singleton
from .class_property import ClassProperty, ClassPropertyMeta
from .protectfile import ProtectFile, ProtectFileError
